#-----------------------------------------------------------------------------
#
# Copyright 2006-2008 Vendmr, daltxguy
# Based on CutList.rb, Copyright 2005, CptanPanic
#
# Permission to use, copy, modify, and distribute this software for
# any purpose and without fee is hereby granted, provided something the above
# copyright notice appear in all copies.
#
# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#-----------------------------------------------------------------------------
#
# Name        : CutListAndMaterials.rb
#
# Type        : Tool
#
# Description : Makes a cutlist based on all selected components in model.
#
# Menu Item   : Plugins -> CutList
# Context-Menu: None
#
# Authors      : <michael robinson> vendmr@yahoo.com 
#                   <steve racz> daltxguy@gmail.com
#
# Currently maintained by :  <steve racz> daltxguy@gmail.com
#
# Usage       : Note: the following description higlights some of the key features. Help is available on each item
#                  of the configuration menu once the cutlist plugin has been started.
#                Call script using Plugins menu, and file name *-CutList.cvs is created in the model's
#               folder.  It is readable by text editors, and cutlist programs like CutList Plus.
#               This script uses bounding boxes to ascertain the dimensions of parts within the user
#               selection.  Only components, groups and components within components are recorded.  Ensure that
#               the component axes are adjusted to the component to give the smallest bounding box to
#               get accurate sizes.  A component which has been thusly created and then rotated will be correctly calculated
#               based on its original axes aligned bounding box. ( In other words, the best way to create a part which is not axes
#               aligned is to create one which is and then rotate it )
#               Volumes are given in board foot or in cu.m. depending on the units used for the model. You can force board 
#               volumes to be in board feet even if the model is in metric units using the 'force board feet' option. This is useful
#               for using metric to build the model but to order wood when the local market ueses imperial.
#               Any component whose name contains the string Part or part will not be treated as wood
#               for the cutlist but as a hardware part and it will be counted.
#               Any component that has a material whose name contains the word sheet or Sheet will
#               be treated as sheet material in a separate list.  Sheet material is calculated in
#               square feet rather than board feet.
#
# Installation: Instructions are available at http://lumberjocks.com/jocks/daltxguy/blog/5143
#
# Additional info: Available at http://steveracz.com/joomla/content/view/45/1/
#
# Date last modified        : Mar 2009
#
# Version     :        1.0                        Initial Release.
#             : 2.0         Added recognition of hardware parts
#             : 2.5         Works with groups, recognizes sheet material
#             : 2.6         Changed lists to compact display with quantity column
#             : 3.0         Added a web user interface, major code refactor, CutListPlus support,
#                           metric support
#             : 3.1         Fixed web display of feet.  Unique names given to unnamed components.
#             : 3.2         Added total length for components in compact view.
#             : 3.2a        Fixed CutListPlus output, Add Force Bd Ft measure ( overrides volume in bd ft even if metric measure used )
#             : 3.3         Added nominal size support.
#             : 4.0        Added new functionality to layout boards. Major code refactor.
#                           Added 'select all components' request if nothing selected by user, Restructured and tidied classes and ruby code
#                            Added tabs to the config html. Renamed name which appears on plugin menu and nested items. Make parts and sheet keyword searches case insensitive. 
#                            Add syntax to exclude words from the search. Fix html error with material names with '<','>' characters. Cutlist solid parts and
#                             sheet parts are now objects instead of arrays. All Cutlist part measures are stored in inches and converted only when displaying
#                             Float class extended to introduce simple syntax for rounding. Rounding is only done on output.
#                             Now displays (or prints) only non-nil lists ie:if no sheet parts found, then only components and parts are displayed etc.
#             : 4.0.1       Add support for SVG output for the layout.
#             : 4.0.2      Add support for Mac devices - change forward slashes to backslashes for file paths, eliminate \n
#             : 4.0.3      Add support for scaled components and groups.
#             : 4.0.4      Fix for layout output when only sheet good components are present.
#             : 4.0.5      Split SVG layout into multiple files when layout spans multi pages. Fix mpath to be the directory where the model is located so files
#                            are written to a known location ie: the same diretory as the model.
#             : 4.0.6      Fix accuracy of metric volume. Change cu.cm to cu.m - a more common measure in metric.
#                            In non-english world, measures are written  as 1.000,00 compared with 1,000.00. Ie: meaning of decimals/commas
#                            are reversed, therefore CSV files need to have a different delimiter than commas, use ";" instead.
#                            Add dedicated output driver for CutList PLus to better match the import requirements for CutList Plus 2009
#             : 4.0.7      Improve SVG pagination. 4x8 sheets come out in separate pages now.
#                            Hardware parts on now matched on either their material name or their part name using the 'part words'.
#                            Fixed bug in component selection: If a part appeared in a nested component, sometimes the top level component would get included in the cutlist
#                            Layout only displays selected set of components now. Previously the component/sheet 'type' setting was ignored.
#                            Add model name to the cutlist output and each page of the layout.
#                            'Select ALL' when prompted by the plugin, ( if nothing was selected) will now select only the visible components, allowing layers to be excluded from the cutlist
#                               In addition, the getSubComponents parser will select only visible sub-components
#                            Minor field naming cleanup for partwords/sheetwords in cutlist.html. No functional change.
#
#-----------------------------------------------------------------------------

require 'sketchup.rb'

# Some utility definitions of a general scope

# determine amount of debugging output to the ruby console
$verbose1 = true # minimal progress tracking
$verbose = false # the whole enchilada - slows down processing considerably - use sparingly.
$verboseComponentDiscovery = false # trace model entity list traversal only

# add a method to return if the current model units is metric
def metricModel? 
  model = Sketchup.active_model

  # Get the length Units of the active model from the unitsOptions
  # 0=inches,1=feet,2=mm,3=cm,4=m
  unit = model.options["UnitsOptions"]["LengthUnit"]
  return !(unit==0 || unit==1)
end
  
def modelInMeters?
  model = Sketchup.active_model
  unit = model.options["UnitsOptions"]["LengthUnit"]
  return (unit==4)
end
  
# extend class Float to provide rounding to x digits
class Float
  def round_to(x)
    (self * 10**x).round.to_f / 10**x
  end
end

# extend class string to be able to convert the string to an html compliant string
# as well as some manipulations required for CutList Plus compatibility
class String
#This will html-ise a string so that we don't have problems displaying in html
# returns a copy of the string with problematic characters replaced by escape sequences
  def to_html
    val = self.gsub(/[']/, "\\\\\'")
    val = val.gsub(/["]/, "&quot;")
    val = val.gsub(/[<]/, "&lt;")
    val = val.gsub(/[>]/, "&gt;")
    return val
  end 
  
  # cut list plus doesn't like inch character " for inch dimensions on import so remove these
  # feet character ie: "'" is interpreted ok
  # mm characters "mm" are interpreted ok
  # cm characters "cm" are interpreted ok
  # units in m are not allowed, so these must be converted prior to this
  def to_clp
    val = self.gsub(/["]/,"")
    return val
  end
end

#extend sketchup class Group so that we can reference the definition from which a 
# group instance has been derived. This makes it analagous to a ComponentInstance
# Sketchup groups also have a component definition but it's not
# directly accessible  so we have to start from the model definitions and search
# looking for the entity which matches ours. Once found we can use it just like
# for Component Instance
class Sketchup::Group
  def definition
    definitions = Sketchup.active_model.definitions
    definitions.each { |definition|
      definition.instances.each { |instance|
        if instance.typename=="Group" && instance == self
          return definition
        end
      }
    }  
    return nil
  end
end

  
#-----------------------------------------------------------------------------
# Reporter class
# This class does the main work of deriving the components from the user selection of the given model, 
# splitting into solid wood components and sheet good components 
# ( and hardware or other 'parts') and then producing the cutlist display in the requested output format
# All methods are private.
# There are two 'entry points', one for sketchup 6 and one for sketchup 5, though, a better description is that there is one
# interface which expects parameters to have been selected externally ( aka sketchup6init) 
# and one where the parameters are a fixed set ( sketchup5init) which only produces the file outputs.
#-----------------------------------------------------------------------------
class Reporter
  
  cutlist_Default_Options = {
    :compactList=> true,
    :listAllSorted => false,
    :linearFeet => true,
    :outFileUsed => true,
    :outFileName => "CutList.csv",
    :clpFileUsed => true,
    :clpFileName => "CutListPlusImport.csv",
    :printPage => false,
    :showComps => true,
    :showSheet => true,
    :showParts => true,
    :forceBoardFeet => false,
    :layout => false,
    :su5 => false,
    :partWords => ["part"],
    :sheetWords => ["sheet"],
    :svgLayout=>false
    }
    
  layout_Default_Options ={
    :fourq => false,
    :fiveq => false,
    :sixq => false,
    :eightq => false,
    :tenq => false,
    :nominalMargin => 0,
    :nominalOut =>false,
    :board4w=>false,
    :board6w=>true,
    :board8w=>false,
    :board10w=>false,
    :board12w=>false,
    :board2l=>false,
    :board4l=>false,
    :board6l=>false,
    :board8l=>true,
    :board10l=>false,
    :board12l=>false,
    :sheet2w=>false,
    :sheet4w=>true,
    :sheet2l=>false,
    :sheet4l=>false,
    :sheet8l=>true,
    :nominalWidth=>true,
    :kerfSize=> 2.inch/16,
    :splitWideParts=>true,
    :splitThickParts=>false,
    :layoutRuleA=>true,
    :layoutRuleB=>true,
    :layoutByMaterial=>false,
    :displayUnplacedParts=>true
  }
  
  @@options = {
  :cutlist_Options => cutlist_Default_Options,
  :layout_Options => layout_Default_Options
  }
  
  ##
  # This entry point method is called by the interactive GUI ( html) configurator
  ##
  def sketchup6Init(cutlist_options, layout_options)
    # merge user selected options with the defaults to get the working set for this session
    @@options[:layout_Options].merge!(layout_options)
    @@options[:cutlist_Options].merge!(cutlist_options)
    
    # determine the flavor of the model, metric or inches
    @metric = metricModel?
    
    # sets whether or not we ask for permission to update group instances found in the model he first time we encounter 
    # a case of it. Leave off for now. This means
    # that it won't ask and so permission is never granted. This feature might be useful but potentially confusing because
    # it alerts to a sketchup internal condition and the user may not realize the implications of answering one way or another.
    # (If you modify a group instance, it automatically is made unique so that the group definition no longer applies to it. If you then
    # wanted to make a change to the group definition, you would have to make the same change to ALL instances, thus losing some
    # of the advantages of it being in a group in the first place.
    @askFirstTime = false
    
    # Determine the units to use for volume measurements in the output
    # If metric units selected, then use metric volume measurements ( ie cu.m.)
    # If metric units selected but forceBoardFeet is on, then use board feet
    # If Imperial measure, then always use board feet
    @volumeMeasureInMetric = (@metric && !@@options[:cutlist_Options][:forceBoardFeet])

    # establish specific sets of options, based on layout and cutlist options, one for boards and one for sheets
    setBoardOptions()
    setSheetOptions()
    
    # create an input board list - empty for now until we create an interface for it
    @inBoardList = BoardList.new()
                       
    # do the work of the reporter class
    main()
  end # def Reporter::sketchup6Init
  
  def setBoardOptions
    setBoardThicknessOptions
    setBoardWidthOptions
    setBoardLengthOptions
    @boardOptions = [@boardLengthOptions,@boardWidthOptions,@quarter, @@options[:layout_Options][:nominalWidth]]
  end

  def setSheetOptions
    setSheetWidthOptions
    setSheetLengthOptions
    # sheets always use nominalWidth, so set that option true always
    # sheets don't use quarter options, so set up a dummy one for sheets with all false
    @sheetQuarter = [false,false,false,false,false]
    @sheetOptions = [@sheetLengthOptions,@sheetWidthOptions,@sheetQuarter,true]
  end
  
  def setBoardThicknessOptions
    #produce an array with all four options placed in it in order
    @quarter = [ @@options[:layout_Options][:fourq], 
                       @@options[:layout_Options][:fiveq], 
                       @@options[:layout_Options][:sixq], 
                       @@options[:layout_Options][:eightq],
                       @@options[:layout_Options][:tenq] ]
  end
  
  def setBoardWidthOptions
    # produce an array with the different widths available
    # this construct makes it easier to iterate over the available options
    # convert lengths, which are in feet, to inches, so that all units are the same internally.
    @boardWidthOptions = Array.new
    @boardWidthOptions.push(4) if @@options[:layout_Options][:board4w]
    @boardWidthOptions.push(6) if @@options[:layout_Options][:board6w]
    @boardWidthOptions.push(8) if @@options[:layout_Options][:board8w]
    @boardWidthOptions.push(10) if @@options[:layout_Options][:board10w]
    @boardWidthOptions.push(12) if @@options[:layout_Options][:board12w] 
  end
  
  def setBoardLengthOptions
    # produce an array with the different lengths available
    # this construct makes it easier to iterate over the available options
    @boardLengthOptions = Array.new
    @boardLengthOptions.push(2*12) if @@options[:layout_Options][:board2l]
    @boardLengthOptions.push(4*12) if @@options[:layout_Options][:board4l]
    @boardLengthOptions.push(6*12) if @@options[:layout_Options][:board6l]
    @boardLengthOptions.push(8*12) if @@options[:layout_Options][:board8l]
    @boardLengthOptions.push(10*12) if @@options[:layout_Options][:board10l]
    @boardLengthOptions.push(12*12) if @@options[:layout_Options][:board12l] 
  end

  def setSheetWidthOptions
     @sheetWidthOptions = Array.new
     @sheetWidthOptions.push(2*12) if @@options[:layout_Options][:sheet2w]
     @sheetWidthOptions.push(4*12) if @@options[:layout_Options][:sheet4w]
  end

  def setSheetLengthOptions
     @sheetLengthOptions = Array.new
     @sheetLengthOptions.push(2*12) if @@options[:layout_Options][:sheet2l]
     @sheetLengthOptions.push(4*12) if @@options[:layout_Options][:sheet4l]
     @sheetLengthOptions.push(8*12) if @@options[:layout_Options][:sheet8l]
  end
  
  # does the main work of the reporter class ie: this is the mainline for the cutlist plugin
  def main    
    # derive the component set required for cutlist from this model. Method components() establishes the working cutlist database
    if ( components() != nil )
      # produce a layout if requested. Layout() creates the necessary database for layout output based on the cutlist database
      if(@@options[:cutlist_Options][:layout] || @@options[:cutlist_Options][:svgLayout] )
        layout()
      end
      # produce the requested cutlist output in the requested formats
      output()
    end
  end
  
  def layout
      # Note:We reference the show parts and show sheet parts options here instead of on the output for efficiency. cutting down on the
      # generation of layout speeds up the output. The output will just display whatever was generated here.
      # determine if we need to create a layout for solid parts
      @unplacedPartsList = CutListPartList.new
      if !@solidPartList.empty? && @@options[:cutlist_Options][:showComps]
        # Split the list of parts into categories to be used for layout based on the selection options
        # make a copy of the partLists object because the layout engine is destructive
        solidPartList = @solidPartList.deep_clone
        solidPartPreParser = SolidPartPreParser.new(solidPartList,@inBoardList,@boardOptions,@@options[:layout_Options],@volumeMeasureInMetric)
        @listOfSolidPartsLists, unplacedPartsList = solidPartPreParser.run
        @unplacedPartsList +(unplacedPartsList)
        
        # for each category of solid parts, layout on as many boards as required
        @layoutBoards = Array.new
        @listOfSolidPartsLists.each { |solidPartList|
          solidPartLayoutEngine = BestFitLayoutEngine.new(solidPartList,@inBoardList,@boardOptions,@@options[:layout_Options],@volumeMeasureInMetric)
          layoutBoards, unplacedPartsList = solidPartLayoutEngine.run
          @layoutBoards += layoutBoards if layoutBoards != nil
          @unplacedPartsList +(unplacedPartsList)
        }
      end
      # determine if we need to create a layout for sheet parts
      if !@sheetPartList.empty? && @@options[:cutlist_Options][:showSheet]
        # Split the list of sheet parts into categories to be used for layout based on the selection options
        sheetPartList = @sheetPartList.deep_clone
        sheetPartPreParser = SheetPartPreParser.new(sheetPartList,@inBoardList,@sheetOptions,@@options[:layout_Options],@volumeMeasureInMetric)
        @listOfSheetPartsLists,  unplacedPartsList = sheetPartPreParser.run
        @unplacedPartsList +(unplacedPartsList)
        
        # for each category of sheet parts, layout on as many boards as required
        @layoutSheets = Array.new
        @listOfSheetPartsLists.each { |sheetPartList|
          sheetPartLayoutEngine = BestFitLayoutEngine.new(sheetPartList,@inBoardList,@sheetOptions,@@options[:layout_Options],@volumeMeasureInMetric)
          layoutSheets, unplacedPartsList = sheetPartLayoutEngine.run
          @layoutSheets += layoutSheets if layoutSheets != nil
          @unplacedPartsList +(unplacedPartsList)
        }
      end
  end
        
  ### Get the material assigned faces within a component if it is not assigned at
  ### the component level
  def getMaterial(component)
    bits = nil
    if(component.typename == "ComponentInstance")
      bits = component.definition.entities
    elsif(component.typename == "Group")
      bits = component.entities
    end ##if

    for f in bits
      if f.typename == "Face"
        materialClass = f.material
        if(materialClass!=nil)
          return materialClass.name
        end
      end
    end
    return "Not assigned"
  end ##getMaterial

  # check if this component is a group with an instance definition
  # if so, then we can derive a name from the original definition if there is one  
  def getGroupCopyName(entity)
    name = ""
    definitions = Sketchup.active_model.definitions
    definitions.each { |definition|
      definition.instances.each { |instance|
        if instance.typename=="Group" && instance == entity
          #now go through this definition and see if there is an instance with a name, return it if found
          definition.instances.each { |i|
            if ( i.name != "" )
              name = i.name
              # now let's do it again but actually set the name of all instances to the one found if user oks this
              if @askFirstTime
                if ( UI.messagebox("Copied group parts found with no name. Ok to set to the same name as the master copy?",  MB_OKCANCEL) == 1 )
                  @okToCopyName = true
                else
                  @okToCopyName = false
                end
                @askFirstTime = false
              end
              if @okToCopyName
                definition.instances.each { |i| i.name = name }
              end
              break
            end
          }
          return name
        end
      }
    }    
    return name
  end
  
  #-----------------------------------------------------------------------------
  # The getSubComponent method is called recursively to derive the fundamental component parts of the model from all
  # components and groups in the selection.
  # returns true if a component level has components
  # Adds components to our component data base as it encounters them and sorts them into hardware parts, sheet goods or solid goods
  # If a component has no sub-components below it, then it must be an elemental part and it is added.
  # Top level components are not added if they have sub-components ( this is because the bounding box of the 
  # top level components encompasses all of its subcomponents and leads to incorrect cutlists)
  # Top level components have sub-components if it or any of its sub-components have sub-components. In other words
  # sub-componentness is transitive.
  # Basically this is a search for all of the basic components. The only exception is for hardware parts where we
  # return the highest level ( on the assumption that the cutlist is not interested in the sub-components of a hardware part)
  # For example, it's not relevent to know that a castor wheel is made up up a wheel, a metal housing, an axle, nuts, etc, at least not in the cutlist context
  #-----------------------------------------------------------------------------
  def getSubComponents(entityList, level)
    puts "checking level=" + level.to_s if $verboseComponentDiscovery
    model = Sketchup.active_model
    selection = model.selection
    # the levelHasComponents flag is used to indicate if we have found any parts at this level
    levelHasComponents  = false
    for c in entityList
      inSelection = selection.contains? c
      #Sub components do not appear as part of the selection so let them through but only look at visible sub-components
      if ( (inSelection || level>1) && c.layer.visible?)

        if c.typename == "ComponentInstance" || c.typename == "Group"
          # get the name of the component or group
          compName = nil
          if(c.typename == "ComponentInstance")
            compName = c.definition.name
            puts "component instance with definition name=" + compName.to_s if $verboseComponentDiscovery
          elsif(c.typename == "Group")
            compName = c.name
            puts "group with name=" + compName.to_s if $verboseComponentDiscovery
            if (compName == nil || compName == "" )
              #let's see if this is a copy of a group which might already have a name
              compName = getGroupCopyName(c)
            end
          end ##if
            
          # compare the 'part' words entered by the user to the entity name or to the material name
          # to find the non cutlist parts
          # If this is a hardware part, then we are done with this part
          
            partMaterialClass = c.material
            if(partMaterialClass==nil)
              partMaterial = getMaterial(c)
            else
              partMaterial = partMaterialClass.name
            end

          if (isPartOrSheet( @@options[:cutlist_Options][:partWords], partMaterial ) ||
              isPartOrSheet( @@options[:cutlist_Options][:partWords], compName) )
            @partList.add(compName)
            puts "adding part name=" + compName.to_s + " level=" + level.to_s  + " as a hardware part since material or name matched"  if $verboseComponentDiscovery
            puts "+++++++++++++++++++++++++++" if $verboseComponentDiscovery
            # since a part was added, mark this level as having components
            levelHasComponents = true
            next   #move on to the next part at this level
          end 

          subList = nil
          if(c.typename == "ComponentInstance")
            subList = c.definition.entities
          elsif(c.typename == "Group")
            subList = c.entities
          end ##if
          
          # go  one level deeper if we found a type of part that might have subparts which we want to add to our list
          hasSubComponents = getSubComponents(subList, level+1) #if check
          if (!hasSubComponents )
           puts "adding part name=" + compName.to_s + " level=" + level.to_s  + " since level=" + (level+1).to_s + " has no subcomponents" if $verboseComponentDiscovery
           puts "+++++++++++++++++++++++++++" if $verboseComponentDiscovery
            ### allows names with - + at start etc
            name = " "+compName
            
            ### If no name is given generate one based on size so that same size unnamed object get grouped together.
            if(name== " ")
              name = "noname"
            end 

            materialClass = c.material
            if(materialClass==nil)
              material = getMaterial(c)
            else
              material = materialClass.name
            end

            # compare the 'sheet' words entered by the user against the material name
            # if there is a match then this selected entity becomes a sheet good object
            # Everything else is a solid part 
            if ( isPartOrSheet( @@options[:cutlist_Options][:sheetWords], material ) ||
                 isPartOrSheet( @@options[:cutlist_Options][:sheetWords], name ) )
              sheetPart = SheetPart.new(c,name,material,@volumeMeasureInMetric)
              # add to the list
              @sheetPartList.add( sheetPart )
            else
              solidPart = SolidPart.new(c,
                                                 name,material, 
                                                 @@options[:layout_Options][:nominalMargin], 
                                                 @quarter,
                                                 @@options[:layout_Options][:nominalOut], 
                                                 @volumeMeasureInMetric)
              # add to the list
              @solidPartList.add( solidPart )
            end  ##if
          else
            puts "skipping partname=" + compName.to_s + " at level=" + level.to_s  + " since level=" + (level+1).to_s + " has subcomponents" if $verboseComponentDiscovery
            puts "--------------------------" if $verboseComponentDiscovery
          end
          # if the level below had no subcomponents, then we just added this part at this level, so mark this level as having components
          # if the level below us had subcomponents, then so must this one by transitiveness, even if none specifically
          # existed at this level ( there could be nested top level components), so in either case we set the level to have components
          levelHasComponents = true
        end
      end#if
    end#for c
    puts "returning levelHasSubcomponents=" + levelHasComponents.to_s + " for level=" + level.to_s if $verboseComponentDiscovery
    return levelHasComponents
  end #getSubComponents

  #-----------------------------------------------------------------------------
  #  Checks to see if component is a Part or Sheet 
  # Make this so it is not case sensitive, so you don't have to enter all possibilities of the same word, capitalized and not
  # Also support special characters similar to Google search to allow specific exclusion of words
  # "-" character ahead of the word means do not include matches on this word ( for example -partition for a part word
  # means that a component named 'partition' should not be considered a part but rather a component which should be 
  # included in the cutlist).
  # The '-' must immediately precede the word, no spaces
  #-----------------------------------------------------------------------------
  def isPartOrSheet( inList, compName )
    @found = false
    @exclusionFound  = false
    inList.each do |p|
      matchWord = p
      
	    # check for the exclusion syntax ( a negative in front of the word)
      exclude = ( (p =~ /^-/)  != nil )
      
      # if  nothing follows the -, then ignore this list word - wrong syntax
      next if ( $' == '' && exclude )
      
      # if the '-' matches the first character, use the part which didn't match as the search string
      matchWord = $' if exclude
      
      # see if the list word matches anywhere in the component name  - case insensitive
      if ( (compName.index(/#{matchWord}/i)) != nil  )
          # Exclusions trump inclusions no matter where they are placed in the list
          if ( exclude )
            @exclusionFound = true
            @found = false
          end
          @found = true if ( !@exclusionFound )
      end
    end # end do loop
    # return the result of the search through all the words. 
    return @found
  end ## end isPartOrSheet
  
  #-----------------------------------------------------------------------------
  # Invert the current selection
  # Takes a current model, changes the selection to be the complete inverse of the
  # current selection and returns the modified model
  #-----------------------------------------------------------------------------
  def invert_selection!(model)
        ss = model.selection
        model.entities.each {|e| ss.toggle(e)}
        return model
  end

  #-----------------------------------------------------------------------------
  # Determine the selection from the model ( or force select all if none was selected)
  # and then decompose the selection to a list of components to be included in cutlist,
  # sheet goods and other parts
  # Either pops up an error to the user and returns or if components found, 
  # when done, there is a list of components in @solidPartlist, sheet goods in @sheetPartlist
  # and parts in the @partList
  #-----------------------------------------------------------------------------
  def components    
    # create a new parts list
    @partList = PartList.new()
    
    # create the new Solid part list
    @solidPartList = SolidPartList.new()
    
    # create the new Sheet part list
    @sheetPartList = SheetPartList.new()
    
    # select the current model
    model = Sketchup.active_model
    
    # get the parts of the model selected
    selection = model.selection
    # start undo...
    model.start_operation "undo"

    # If the current selection is empty, then assume that the entire  model is to be selected.
    # toggle the selection to select all, if still empty, then  display a message
    if ( selection.empty? )
      # try selecting all
      # confirm with the user that this is ok
      if ( UI.messagebox("Nothing was selected from the model.\nDo you want to select all visible? ",  MB_OKCANCEL) == 1 )
        
        # inverse the empty selection to select all
        model = invert_selection!(model)
        @selection_inverted = true

        # get the selection from the model again
        selection = model.selection
        
        #remove any entities from the selection which are not visible
        selection.each { |entity| selection.toggle( entity) if !entity.layer.visible? }
        
        # if it's still empty, then there must be nothing in the model.
        if ( selection.empty? )
          UI.beep
          UI.messagebox("Your model is empty or no entities are visible.\nNo Cutlist generated.")
          return nil
        end
      else
        # user cancelled from the select all request
        return nil
      end
    end

    entities = model.entities
    @mname = model.title

    # check model has a directory path, so we know where to store the output 
    mpath = model.path
    puts "Model path=" + model.path.to_s
    if mpath == ""
      UI.beep
      UI.messagebox("You must save the 'Untitled' new model \nbefore making a Component Report !\nNo Cutlist generated.")
      return nil
    end
     
    # now get the actually directory from the path, so we can put our files in the same directory.
     
    @mpath = File.dirname(mpath)
    puts "directory= " + @mpath.to_s

    ### show VCB and status info...
    Sketchup::set_status_text(("CUT LIST GENERATION..." ), SB_PROMPT)
    Sketchup::set_status_text(" ", SB_VCB_LABEL)
    Sketchup::set_status_text(" ", SB_VCB_VALUE)

    #main work of deriving the components from the selection of the model. This updates @solidPartList and @sheetPartList the components and sheet good lists respectively
    puts "Component Discovery start ---->" if $verboseComponentDiscovery
    getSubComponents(entities,1)
    puts "Component Discovery end <----" if $verboseComponentDiscovery
    
    # if no components selected or no parts then exit...
    if ( @solidPartList.empty? && @sheetPartList.empty? && @partList.empty?)
      UI.beep
      UI.messagebox("No Components found in your model.\nYou must create a Component from your selection.\nClick on Help for more info.\nNo Cutlist will be generated.")
      return nil
    end
    
    #finally sort the solid component list and sheet list if the option was selected
    if @@options[:cutlist_Options][:listAllSorted]
      @solidPartList.sort
      @sheetPartList.sort
    end
    
        # commit undo...
    model.commit_operation

  end # components
  
def output
    ### HTML output ###
    if(@@options[:cutlist_Options][:printPage])
      cutlistHtml = HtmlOutputDriver.new(@@options[:cutlist_Options][:compactList],
                                                       @@options[:cutlist_Options][:showComps],
                                                       @@options[:cutlist_Options][:showSheet],
                                                       @@options[:cutlist_Options][:showParts],
                                                       @volumeMeasureInMetric,
                                                       @solidPartList,
                                                       @sheetPartList,
                                                       @partList,
                                                       @mname)
      cutlistHtml.run
    end ## if

    ### File output ###
    if(@@options[:cutlist_Options][:outFileUsed])
      cutlistCsvFile = FileOutputDriver.new(@@options[:cutlist_Options][:compactList],
                                                         @@options[:cutlist_Options][:showComps],
                                                         @@options[:cutlist_Options][:showSheet],
                                                         @@options[:cutlist_Options][:showParts],
                                                         @volumeMeasureInMetric,
                                                         @solidPartList,
                                                         @sheetPartList,
                                                         @partList,
                                                         @mname,
                                                         @mpath,
                                                         @@options[:cutlist_Options][:outFileName])
      cutlistCsvFile.run
    end ## if

    ### CutListPlus output ###
    if(@@options[:cutlist_Options][:clpFileUsed])
      cutlistClpFile = ClpFileOutputDriver.new(@@options[:cutlist_Options][:compactList],
                                                            @@options[:cutlist_Options][:showComps],
                                                            @@options[:cutlist_Options][:showSheet],
                                                            false,
                                                            @volumeMeasureInMetric,
                                                            @solidPartList,
                                                            @sheetPartList,
                                                            @partList,
                                                            @mname,
                                                            @mpath,
                                                            @@options[:cutlist_Options][:clpFileName])
      cutlistClpFile.run
    end ## if
    
    if(@@options[:cutlist_Options][:layout])
      @unplacedPartsList = nil if !@@options[:layout_Options][:displayUnplacedParts]
      cutlistLayoutHtml = HtmlLayoutDriver.new(@layoutBoards,
                                                               @layoutSheets,
                                                               @unplacedPartsList,
                                                               @mname)
      cutlistLayoutHtml.run
    end
    
    if(@@options[:cutlist_Options][:svgLayout])
      @unplacedPartsList = nil if !@@options[:layout_Options][:displayUnplacedParts]
      cutlistSvgLayoutHtml = SvgLayoutDriver.new(@layoutBoards,
                                                                    @layoutSheets,
                                                                    @unplacedPartsList,
                                                                    @mname,
                                                                    @mpath)
      cutlistSvgLayoutHtml.run
    end

  end#def output

end#class Reporter


#-----------------------------------------------------------------------------
# module PartsIntelligence
# Used to answer any questions we may have about the partslist and
# any board options which may help us in selecting boards, making or layout
# optimising the layout or maybe even determining that the layout is not possible
#-----------------------------------------------------------------------------
module PartsIntelligence
  # various reasons why layout will not converge or is impossible
  def layoutImpossible?
    return true if @inPartList.empty?
    return true if !@homerDepot.boardsAvailable?
  end
  # some queries about the partslist which will help us decide
  # how to lay it out
  def allPartsSameThickness?
    @inPartList.allPartsSameThickness?
  end
  def allPartsSameMaterial?
    @inPartList.allPartsSameMaterial?
  end
  # do I have any parts which are wider than the available boards?
  def partsWiderThanBoards?
    @widestBoard = @homerDepot.widestBoardAvailable?
    puts "widest Board Available=" + @widestBoard.to_s if $verbose1
    partArray = @inPartList.getList
    partArray.each { |part| return true if part.getWidth > @widestBoard }
    return false
  end
  
  # do I have parts which are currently thicker than the thickest boards available?
  def partsThickerThanBoards?
    @thickestBoard = @homerDepot.thickestBoardAvailable?.inch
    puts "thickest Board Available=" + @thickestBoard.to_s if $verbose1
    partArray = @inPartList.getList
    partArray.each { |part| return true if part.getThickness > @thickestBoard }
    return false
  end
  
  # do I have parts which require thicker boards than what is available?
  def partsThinnerThanBoards?
    @thickestBoard = @homerDepot.thickestBoardAvailable?.inch
    partArray = @inPartList.getList
    partArray.each { |part| return true if part.getThickness < @thickestBoard }
    return false
  end
  
  # are many parts a multiple of the width of one of the baords
  # if so, then return that width and we'll prefer this size board for layout
  # after splitting the parts into mutiple sections
  def partsMutipleofWidth?
  end
  # determines if complete layout is not possible if the sq footage of the parts
  # exceeds the available boards
  def morePartsThanBoards?
    return false
  end
end

#-----------------------------------------------------------------------------
# The LayoutPreParser takes the parts list and the options and organizes
# the parts by material and thickness, ready for layout
#-----------------------------------------------------------------------------
class LayoutPreParser
  # add mixin module to answer any questions we may have which will help us place
  # the parts
  include PartsIntelligence
  
  def initialize(partsList, inBoardList, partOptions,layoutOptions,metricVolume)
    @homerDepot = BoardSuperMarket.new(inBoardList, partOptions,metricVolume)
    @layoutOptions = layoutOptions
    @layoutByMaterial = layoutOptions[:layoutByMaterial] 
    @quarters = partOptions[2].include? true
    @listOfPartsLists = Array.new()
    @listOfPartsLists.push(partsList)
    @inPartList = partsList
    @status = LayoutStatus.new("","success")
    @partsIgnoreList = CutListPartList.new()
  end
  
  # The preParser makes use of the intelligence module which is a series
  # of methods to answer various questions which may assist in the decision
  # making to better or faster the decision making process and therefore
  # a convergence to an optimal layout (as optimal as the rules we have coded allow for
  # of course)
  def partsPreParser
    puts "partsPreParser:Looking for splits needed..."
    if layoutImpossible?
      @status.set("No parts or no boards.\nLayout is not possible","error")
      return
    end
    if morePartsThanBoards?
      @status.set("Parts exceeds available boards.\nLayout is not possible","error")
      return
    end
    # if no quarters options were given, then skip this step and keep all parts thicknesses as originally created
    # otherwise split any parts up which are thicker than the given board sizes and make them fit on these thicknesses
    if ( partsThickerThanBoards? || partsThinnerThanBoards? )&& @quarters
      # I can lay these out only if you have given me permission to split these into glue-ups
      if @layoutOptions[:splitThickParts]
        # First adjust parts which are still thinner than the thickest board
        # This can occur if the margin specified by the user  + part thickness exceeded the available board thicknesses
        # in this case the part thickness was left unchanged, but since for layout we are allowed to split the parts
        # we can just make the part as thick as it needs to be and let the next step split it if necessary
        normalizeThinGlueUpParts(@thickestBoard)
        # This next step now takes parts thicker than the thickest board
        normalizeGlueUpParts(@thickestBoard)
      else
        @status.set("Some parts require thicker boards than is available.\n" + \
                          "Either check the 'Split Thick Parts' option or specify" + \
                          " thicker boards\n" + \
                          "Some or all of the parts will not be layed out","warning")
        removeThickParts(@thickestBoard)
      end
    end
    if !allPartsSameMaterial?
      # we'll have to split into multiple lists and get different boards for each list
      # we actually make this decision later
    end
    if partsWiderThanBoards?
      # we'll have to split the parts up to lay them out on the available boards
      # split parts up based on the widest board available. 
      # Note this may be a simplistic decision if there are real boards but only some wides ones and 
      # splitting it up this way may still result in the parts not being placed. 
      # Since our list of parts and boards are sorted from largest to smallest, the hope is that we'll find a spot for them all
      # If not, then our todo list will contain boards not placed. It may turn out that splitting the boards in a different fashion
      # would have made them fit, but this is much more difficult decision to make here, so we will leave that to a future version
      # I suppose some sort of adaptive method must be used. If too many parts left over, try a different decision and start again
      # and return the best of the tries  - an interesting problem to solve!
      
      normalizePartsList(@widestBoard) if @layoutOptions[:splitWideParts]
    end
  end
  
  def removeThickParts(thickestBoard)
    partArray = @inPartList.getList
    removalList = Array.new
    partArray.each { |part| 
      if part.getThickness.inch > thickestBoard.inch
        removalList.push(part)
      end
    }
    #now adjust the parts lists - move from main part list to ignore list
    removalList.each{ |oldPart| 
      @inPartList.remove(oldPart)
      @partsIgnoreList.add(oldPart)
     }
end

  # This will split parts which are too thick into as many pieces of the given thickness required to glue-up this part
  # Limit the number of pieces. Anything that has more than say 10 glue ups will be ignored
  def normalizeGlueUpParts(thickestBoard)
    partArray = @inPartList.getList
    removalList = Array.new
    addList = Array.new
    partArray.each { |part| 
      #puts "part=" + part.getName + " index=" + partArray.index(part).to_s
      if part.getThickness.inch > thickestBoard.inch
        # need to figure out how many chunks to split it into
        splitInto = (part.getThickness.inch/thickestBoard.inch).ceil
        if splitInto >= 10
          # this piece is too big for us, reject it
          @status.set("Part " + "'" + part.getName + "'" + " is too thick" + "\nThis part will not be split and will be ignored in the layout ","warning")
          removalList.push(part)
          @partsIgnoreList.add(part)
          next
        end
        # The thickness of each part will always be the thickness of the board
        splitThickness = thickestBoard.inch
        puts "part=" + part.getName + " thickness=" + part.getThickness.inch.to_s + " splits=" + splitInto.to_s + "new thickness=" + splitThickness.to_s if $verbose1
        #make a copy(ies) of this part
        splitInto.times do |i| 
          splitPart = part.deep_clone
          splitPart.changeThickness(splitThickness)
          # add the clones to the list
          addList.push(splitPart)
        end
        # mark the original part for removal from the list, it has been replaced with the split parts. Removing it will also adjust the database and delete this thickness bucket
        # if no longer required
        # Note: removing things from the array while traversing it messes up the indices and therefore causes each to skip some entries
        removalList.push(part)
      end
    }
    #now remove the ones marked for deletion
    removalList.each{ |oldPart| @inPartList.remove(oldPart) }
    #add the new parts created
    addList.each{ |newPart| @inPartList.add(newPart)}
    # sort the list again
    @inPartList.sort
    puts @inPartList.inspect if $verbose
 
  end
  
  # This will restore parts to marginal thickness if marginal thickness was too thick for the available
  # boarrd ( and therefore it was left as is ). We do this because we can split this into glueUps when laying out.
  def normalizeThinGlueUpParts(thickestBoard)
    partArray = @inPartList.getList
    removalList = Array.new
    addList = Array.new
    partArray.each { |part| 
      if part.getThickness.inch < thickestBoard.inch
        puts "thin part=" + part.getName.to_s + " Adjust from" +  part.getThicknessString + " to" + part.getMarginThickness.to_l.to_s
        # just adjust this part thickness back to the marginalThickness
        removalList.push(part)
        newPart = part.deep_clone
        newPart.changeThickness(part.getMarginThickness)
        addList.push(newPart)
      end
    }
    #now remove the ones marked for deletion
    removalList.each{ |oldPart| @inPartList.remove(oldPart) }
    #add the new parts created
    addList.each{ |newPart| @inPartList.add(newPart)}
    # sort the list again
    @inPartList.sort
    puts @inPartList.inspect if $verbose
  end
  
  # This will split any parts which are too wide into some multiple of evenly split widths, no wider than the widest board
  # The parts list will be added to in this method and then re-sorted at the end.
  def normalizePartsList(widestBoard)
    partArray = @inPartList.getList
    removalList = Array.new
    addList = Array.new
    partArray.each { |part| 
      if part.getWidth.inch> widestBoard.inch
          # need to figure out how many chunks to split it into
          splitInto = (part.getWidth.inch/widestBoard.inch).ceil
          splitWidth = widestBoard.inch
          puts "part=" + part.getName.to_s + " width=" + part.getWidth.to_s + " splits=" + splitInto.to_s + " new width=" + splitWidth.to_s if $verbose1
          # change the width of the original
          #make a copy(ies) of this part, 1 less than the number of splits
          splitInto.times do |i| 
            splitPart = part.deep_clone
            splitPart.changeWidth(splitWidth)
            # add the clones to the list
            addList.push(splitPart)
          end
          # mark the original part for removal from the list, it has been replaced with the split parts. 
          removalList.push(part)
      end
      }
    #now remove the ones marked for deletion
    removalList.each{ |oldPart| @inPartList.remove(oldPart) }
    #add the new parts created
    addList.each{ |newPart| @inPartList.add(newPart)}
    # sort the list again
    @inPartList.sort
    puts @inPartList.inspect if $verbose
  end
  
  # organize the parts by layout board type, readying for the layout
  def partsSplitter
    if @layoutByMaterial  
        @listOfPartsLists = @inPartList.splitPartsListByThicknessAndMaterial
    else
        @listOfPartsLists = @inPartList.splitPartsListByThickness
    end
  end
  
  def run
    # gather some useful analysis of the parts list so as to make more intelligent choices of which
    # parts to place on which boards or what kind of boards to ask for
    # split wide or too thick pieces into standardized sizes if user has requested this
    partsPreParser()

    return nil if @status.statusNotGood
    
    # split into multiple lists by materials or thickness or both depending on user options
    # one list per layout type ie: all 3/4" parts on Cherry
    partsSplitter()
    
    return nil if @status.statusNotGood
    return @listOfPartsLists, @partsIgnoreList
  end
end

class SolidPartPreParser < LayoutPreParser
end

class SheetPartPreParser < LayoutPreParser
end

#-----------------------------------------------------------------------------
# Class BoardWarehouse is an abstraction of all information about boards or sheet goods 
# which are available to use for the layout.
# These may be actual boards which the user has provided us or it is a 
# description of boards which may be used.
# It manages getting a board closest to your specification either inventing one ( ordering it
# presumable from some sustainably managed pool of electronic boards) or getting one out
# of the warehouse ( which the user has obviously obtained from a sustainable source)
# Boards may be returned for a different size if it didn't suit and substituted for another one
#-----------------------------------------------------------------------------
class BoardWarehouse
  def initialize(metricVolume)
    @boardList =BoardList.new()
    @boardOptions = [[],[],[],false]
    @metricVolume = metricVolume
  end
    
  # accept a request for a board of a specific size and return the best possible choice, either
  # an actual board, or a 'virtual board', on special order. There are unlimited
  # virtual boards in all sizes but a fixed number of actual boards, otherwise
  # once created, they behave the same way
  def boardOrder(length, width)
  end
  
  # didn't like the board? As long as you didn't cut it, we'll take it back
  # real boards are returned to inventory, whereaas virtual boards just cancels
  # the special order ( ie: nothing changes)
  def boardReturn(board)
  end
  
  # this is the generally used interface. Just returns the best next available board
  # Best here means the largest board available
  # returns an instance of class LayoutBoard
  def getBoard
    #find a board
    #create and return a LayoutBoard
  end
  
  def boardsAvailable?
  end
  
  # determine the widest board in the inventory
  def widestBoardAvailable?
  end
  # determine the thickest board in the inventory
  def thickestBoardAvailable?
  end
end

#-----------------------------------------------------------------------------
# A set of methods for real boards defined as module so it can be mixed in as needed
# to our warehouse class
#-----------------------------------------------------------------------------
module RealBoards
  def getRealBoard(material,thickness)
  end
  
  def getSpecificRealBoard(length,width,thickness,material)
  end
  
  # a delivery of real boards, can also be done after initialization
  def realBoardDelivery(boardList)
    @boardList = boardList
  end
  def anyRealBoardsAvailable?
     !@boardList.empty?
   end
   
  # determine the widest board in the inventory
  def widestRealBoardAvailable?
    return 0 if !anyRealBoardsAvailable?
    boardListArray = @boardList.getList
    widest = boardListArray[0].getWidth
    boardListArray.each {|board| 
      width = board.getWidth;
      widest = width if width > widest
      }
    return widest
  end
  # determine the thickest board in the inventory
  def thickestRealBoardAvailable?
    return 0 if !anyRealBoardsAvailable?
    boardListArray = @boardList.getList
    thickest = boardListArray[0].getThickness
    boardListArray.each {|board| 
      thickness = board.getThickness;
      thickest = thickness if thickness > thickest
      }
    return thickest
  end
end

#-----------------------------------------------------------------------------
# A set of methods for virtual boards (only defined as options)
# defined here as a module so it can be mixed in as needed to our warehouse class
#-----------------------------------------------------------------------------
module VirtualBoards
  # We restrict the number of boards in our inventory to 20 boards mainly as a way of ensuring that
  # the layout doesn't take an excessive length of time. We can change this as the performance of the
  # output improves.
  @@boardLimit = 20
  
  # this method always gets the biggest board possible. If you want a specific board then used
  # getSpecificVirtualBoard
  def getVirtualBoard(material, thickness)
    return nil if !anyVirtualBoardsAvailable?
    # sort legnth and width option arrays
    sortedLength = @boardOptions[0].sort
    sortedWidth = @boardOptions[1].sort
    # get the largest board possible
    longest = sortedLength.last
    widest = sortedWidth.last
    # create a board with these dimensions. 
    # Get the thickness closest to that requested
    newThickness = closestThicknessVirtualBoard(thickness)
    # If nominalWidth asked for,then explicitly get the size requested, otherwise get
    # a dressed board
    if @nominalWidth
      board = Board.new(longest,widest,newThickness,material,@metricVolume)
    else
      board = NominalBoard.new(longest,widest,newThickness,material,@metricVolume)
    end
    # reduce the number of boards in the inventory
    @numberOfBoards -= 1
    return board
  end
  
  # get a board as close to the specification as possible. If nothing is available in the right
  # size, look for the smallest board which is larger. If nothing is larger, get the one which is closest 
  # but smaller. The layout engine will decide if it works or not, we just get what is asked for.
  def getSpecificVirtualBoard(length,width,thickness,material)
    return nil if !anyVirtualBoardsAvailable?
    
    # sort length and width option arrays
    sortedLength = @boardOptions[0].sort
    sortedWidth = @boardOptions[1].sort
    
    #find the closest values in the options table
    closestLength = findClosest(length,sortedLength)
    closestWidth = findClosest(length,sortedWidth)
    
    # wouldn't expect not to find something but just in case
    return nil if (closestWidth == nil || closestLength == nil )
    
    #create a board with these dimensions and return
    if @nominalWidth
      board = Board.new(closestLength,closestWidth,thickness,material,@metricVolume)
    else
      board = NominalBoard.new(closestLength,closestWidth,thickness,material,@metricVolume)
    end
    # reduce the number of boards in the inventory
    @numberOfBoards -= 1
    return board
  end
  
  # takes a number and a sorted array of numbers and finds the entry which is :
  # exactly that number
  # slightly bigger if not exact
  # slightly smaller if nothing bigger or exact
  def findClosest(dimension,optionArray)
    #find all entries larger than or equal to the dimension and pick the first one to return
    largest = optionArray.select{|x| x >= dimension }
    return largest[0] if largest != nil
    #find the one just smaller 
    smallest = optionArray.select{|x| x < dimension }
    return smallest[0] if smallest != nil
    return nil
  end
  
  # delivery via boardOptions methods - can be added after initialization
  def virtualBoardDelivery(boardOptions)
    #not much to do with the options except to store the options
    @boardOptions = boardOptions
    @nominalWidth = boardOptions[3]
    # set number of boards available with this delivery
    @numberOfBoards = @@boardLimit
  end
  
  def anyVirtualBoardsAvailable?
    # have we reached our inventory limit?
    return false if @numberOfBoards == 0
    # if we have any width and length board options, then there are always more boards to be had
    return (!@boardOptions[0].empty? && !@boardOptions[1].empty? )
  end
  
  # determine the widest board in the inventory subject to the user selected option to have undressed boards (nominalWidth)
  def widestVirtualBoardAvailable?
    sortedWidth = @boardOptions[1].sort
    widestBoard = sortedWidth.last
    return widestBoard if @nominalWidth
    return widestBoard -=  1/2.inch  if widestBoard <= 6.inch
    return widestBoard -=  3/4.inch 
  end
  # determine the thickest board in the inventory
  def thickestVirtualBoardAvailable?
    # check which of the "quarter" thickness options has been selected, if any.
    # if none selected, return default of 3/4

    return (2.5).inch if @boardOptions[2][4]
    return 2.inch if @boardOptions[2][3]
    return (1.5).inch if @boardOptions[2][2]
    return (1.25).inch if @boardOptions[2][1]
    return 1.inch if @boardOptions[2][0]
    return 3/4.inch
  end
  
  # find the thickness closest to that requested which is larger
  # if no options set, then return the requested thickness
  def closestThicknessVirtualBoard(thickness)
    return 1.inch if @boardOptions[2][0] && 1.inch >= thickness
    return (1.25).inch if @boardOptions[2][1] && (1.25).inch >= thickness
    return (1.5).inch if @boardOptions[2][2] && (1.5).inch >= thickness
    return 2.inch if @boardOptions[2][3] && (2).inch >= thickness
    return (2.5).inch if @boardOptions[2][4] && (2.5).inch >= thickness
    return thickness
  end
end

# warehouse wth real boards to/from the warehouse
class BoardLumberYard < BoardWarehouse
  
  #include instance methods from the RealBoard module
  include RealBoards
  
  def initialize(boardList,metricVolume)
    super(metricVolume)
    realBoardDelivery(boardList)
  end
  
  def getBoard(material="default", thickness=1.0.inch)
    getRealBoard(material, thickness)
  end
  
  def boardsAvailable?
    anyRealBoardsAvailable?
  end
  
  # determine the widest board in the inventory
  def widestBoardAvailable?
    widestRealBoardAvailable?
  end
  # determine the thickest board in the inventory
  def thickestBoardAvailable?
    thickestRealBoardAvailable?
  end
end

# warehouse with only of board dimension combinations which can be ordered
# ie: boardOptions
class BoardServiceDesk < BoardWarehouse
  
  #include instance methods from the Virtual Board module
  include VirtualBoards

  def initialize(boardOptions,metricVolume)
    super(metricVolume)
    virtualBoardDelivery(boardOptions)
  end
  
  def boardDelivery(boardList,boardOptions)
  end
  
  def getBoard(material="default", thickness=1.0.inch)
    getVirtualBoard(material, thickness)
  end
  
  def boardsAvailable?
    anyVirtualBoardsAvailable?
  end
  # determine the widest board in the inventory
  def widestBoardAvailable?
    widestVirtualBoardAvailable?
  end
  # determine the thickest board in the inventory
  def thickestBoardAvailable?
    thickestVirtualBoardAvailable?
  end
end

# the full service board warehouse
class BoardSuperMarket < BoardWarehouse
  #include instance methods from both the RealBoard module and Virtual Board module
  include RealBoards
  include VirtualBoards
  
  # if you have both options and real boards use this initialization
  def initialize(boardList,boardOptions,metricVolume)
    super(metricVolume)
    realBoardDelivery(boardList)
    virtualBoardDelivery(boardOptions)
  end
  
  def getBoard(material="default", thickness=1.0.inch)
    return getVirtualBoard(material,thickness) if !getRealBoard(material,thickness)
  end
  
  def boardsAvailable?
    return anyVirtualBoardsAvailable? if !anyRealBoardsAvailable?
    return true
  end
  
  def widestBoardAvailable?
    widestRealBoard = widestRealBoardAvailable?
    widestVirtualBoard = widestVirtualBoardAvailable?
    return widestRealBoard if widestRealBoard > widestVirtualBoard
    return widestVirtualBoard
  end
  
  def thickestBoardAvailable?
    thickestRealBoard = thickestRealBoardAvailable?
    thickestVirtualBoard = thickestVirtualBoardAvailable?
    return thickestRealBoard if thickestRealBoard > thickestVirtualBoard
    return thickestVirtualBoard
  end
  
end

#-----------------------------------------------------------------------------
# class Board - raw material from which we can cut the parts
#-----------------------------------------------------------------------------
class Board
  def initialize(length,width,thickness,material,metricVolume)
    # kept internally in inches
    @length = length.inch
    @width = width.inch
    @thickness = thickness.inch
    @material = material
    calcBoardFeet
    
    #store the board size in pixels as well. 100 pixels to the foot
    # as this displays in a reasonable scale.
    # we only need the length and the width
    # width in px represents the actual size of the board
    @length_px = ((@length/12)*100).to_f.round_to(0)
    @width_px = ((@width/12)*100).to_f.round_to(0)
    @area_px = @length_px*@width_px
    @metricVolume = metricVolume
  end
  
  def calcBoardFeet
    @area= @length*@width
    @boardFeet = (@area*@thickness)/144
  end
  
  # print the dimensions of the board
  def to_s
    getWidthString + ' x ' + getLengthString + ', ' + getThicknessString + ' ' + getMaterial + ' ' + getBoardFeetString 
  end
  def getLength
    @length
  end
  def getLengthString
    getLength.to_l.to_s
  end
  def getWidth
    @width
  end
  def getWidthString
    getWidth.to_l.to_s
  end
  def getThickness
    @thickness
  end
  def getThicknessString
    getThickness.to_l.to_s
  end
  def getMaterial
    @material
  end
  def getLengthPx
    @length_px
  end
  def getWidthPx
    @width_px
  end
  def getAreaPx
    @area_px
  end
  def getBoardFeet
    if @metricVolume
      # 1bd ft = 2359.7424/1000000 cu m
      (@boardFeet*(2359.7372232207958956904236222001/1000000)).round_to(4)
    else
      @boardFeet.round_to(2)
    end
  end
  def getRawBoardFeet
    @boardFeet
  end
  def getBoardFeetUnitsString
    return ' cu.m.' if @metricVolume
    return ' bd.ft.' if !@metricVolume
  end
  def getBoardFeetString
    string = getBoardFeet.to_s
    string += getBoardFeetUnitsString
    return string
  end
  
end

#-----------------------------------------------------------------------------
# a nominal board is a board whose actual dimensions are slightly smaller
# than the nominal size ie: 1/2" less when <= 6"  and 3/4" less if > 6"
# board feet is based on the nominal size, not the actual size
# Otherwise everything else about it is the same
# This is the default board we use
#-----------------------------------------------------------------------------
class NominalBoard < Board
  def initialize(length,width,thickness,material,metricVolume)
    @nominalWidth = width.inch
    if @nominalWidth <= 6.inch
      actualWidth = width.inch - 1/2.inch 
    else
      actualWidth = width.inch - 3/4.inch
    end
    super(length,actualWidth,thickness,material,metricVolume)
  end
  def calcBoardFeet
    @area= @length*@nominalWidth
    @boardFeet = (@area*@thickness)/144
  end
  def getWidthString
    return getNominalWidth.to_l.to_s + "(" + getWidth.to_l.to_s + ")"
  end
  def getNominalWidth
    @nominalWidth
  end
end

# used to identify and handle different errors produced by the layout engine
class LayoutStatus
  def initialize(error,type)
    set(error,type)
  end
  def set(error,type)
    @error = error
    @type = type
  end
  def display
    if @type == "fatal"
      UI.beep
      UI.messagebox(@error)
    elsif @type == "error"
      UI.beep
      UI.messagebox(@error)
    elsif @type == "warning"
      UI.beep
      UI.messagebox(@error)
    elsif @type == "choice"
      UI.beep
      UI.messagebox(@error)
    elsif @type == "success"
    end
    #clear the error once displayed
    @error=""
    @type="success"
  end
  def getResultType
    @type
  end
  def good?
    return true if (@type == "" || @type == "success" )
    return false
  end
  def warn?
    return true if (@type == "warning" || @type == "choice" )
    return false
  end
  def stop?
      return true if (@type == "error" || @type == "fatal")
      return false
  end
    
  def statusNotGood
    if good?
      return false
    end
    if warn?
      display
      return false
    end
    if stop?
      display
      return true
    end
    return false
  end
end

#-----------------------------------------------------------------------------
# class LayoutBoard - boards used in the layout - maintains a list of parts
# which are to be cut from this board and the data structure representing 
# used and unused portions of the board.
# Note that this view of the board is specific to this layout algorithm. 
# Other algorithms may need a different LayoutBoardClass
# This is a tree structure. Whenever a part is cut out of the board, the
# remaining area can be represented by two rectangles - two in the horizontal direction
# and two in the vertical direction covering all of the remaining unused space. 
# These two views represent all options for future placement
# and are represented as left and right branches of a tree structure.
# When we go to place the next part, we choose the best option by determining which fills one of
# available spaces the best. When that choice is selected, then we eliminate from the tree the 
# options which are no longer available
# For example, let's say we have a 4x8 sheet of plywood. Place the sheet horizontally.
# if we need a piece cut out 2'x3', then horizontal remaining spaces are 2x5,2x8 ( =10+16=26 sq ft)
# vertical remaining spaces are 3x2 and 5x4 ( =6+20=26bdft)
# As soon as we select one of the 4 choices for the next part, that selects that pair and the pair not chosen
# is immediately eliminated from further choices. Then the part that is placed subdivides the space that we chose
# with 4 options and so on.
# Whenever we go to place a part we explore all available options forone that fits the best ie: with the least amount of waste
# we do this by comparing sq ft of the part to be placed with the sq ft of the available space.
# 
#
#    +------------------+------------------------------------+
#    |                      |                                              |
#    |                      |   2                                         |
#    |___________|                                              |  4
#    |         3                                                          |
#    |                                                                     |
#    |                                                                     |
#    +-------------------------------------------------------+
#                                         8
#-----------------------------------------------------------------------------
  
class LayoutBoard
  class LeafNode
    @@nodeType="leaf"
    def initialize(parent)
      @parent = parent
      @rect = {:xy => Array[0,0], :length => 0, :width => 0 }
    end
    # each leaf node holds a description of the available rectangle on this board
    # This is given by the :xy which is an array of top, left coordinates, relative to 0,0 being at the top left
    # and the :length and :width of the piece in pixels where 12in=100px
    def addRect(rect)
      @rect = @rect.merge!(rect)
      # :xy is the(x,y) coordinate of the top left of the part
      # :length is the length
      # :width is the width
    end
    def rect
      @rect
    end
    def getNodeType
      @@nodeType
    end
    def parent
      @parent
    end
  end
  
  class Node
    @@nodeType="node"
    def initialize(parent)
      @parent = parent
      @left = nil
      @right = nil
      @part = {:xy => Array[0,0], :length => 0, :width => 0 }
    end
    # rect is a description of the space available
    # The parameter rect is a description of the size defined as a pair of (x,y) coordinates
    # where (x1,y1) is the top left and (x2,y2) is the bottom right in a coordinate system
    # where (0,0) is at the top left corner, positive x is from left to right and positive y is from top to bottom
    # left always contains the remaining space described in a  horizontal orientation
    def addLeft(rect)
      @left = LeafNode.new(self)
      @left.addRect(rect)
    end
    # rect is a description of the space available in vertical oreientation
    # see above for how the parameter rect is expected
    def addRight(rect)
      @right = LeafNode.new(self)
      @right.addRect(rect)
    end
    def left
      @left
    end
    def left=(left)
      @left=left
    end
    def right=(right)
      @right=right
    end
    def right
      @right
    end
    def parent
      @parent
    end
    def getNodeType
      @@nodeType
    end
    def part
      @part
    end
    def addPart(part)
      @part = @part.merge!(part)
    end
    def getPartArea
      return @part[:width]*@part[:length]
    end
  end

  def initialize(board,layoutOptions,metricVolume)
    #super(board.getLength,board.getWidth,board.getThickness,board.getMaterial,metricVolume)
    @board = board
    # initialize the list of parts which will be cut out of this board
    @partInBoardList = SolidPartList.new()
    # intialize the root node
    @root = Node.new(self)
    #initial view is that the entire board is represented on one leaf node, left or right makes no diff at this point
    topLeft = Array[0,0]
    @root.addLeft( {:xy =>topLeft,:length =>@board.getLengthPx,:width =>@board.getWidthPx} )
    #@root.addRight( {:xy =>topLeft,:length =>board.getLengthPx,:width =>board.getWidthPx})
    @boardFull = false
    @layoutOptions = layoutOptions
    #kerfSize is in inches. 100Px to 12"
    @kerfSizePx = @layoutOptions[:kerfSize]*(100/12).ceil
  end

  #add a part to be cut out of this board
  #layoutPart is an instance of LayoutPart
  def addPartToBoard(layoutPart)
    # now figure out where it should be placed on the board. Find enough room for the part + kerfsize on one edge of the width and one on the length
    leafNode = findBestFit(@root,layoutPart.getLengthPx + @kerfSizePx,layoutPart.getWidthPx + @kerfSizePx)
    # return if we couldn't fit this part
    return false if leafNode == nil

    # place the locaton info in the parts list as well, against the part it represents.
    layoutPart.addLocationOnBoard(leafNode.rect[:xy])
    
    #add it to our list of parts which are on this board - this is just a list in case we need 
    # quick access later on for display reasons or counting, without having to traverse the tree
    @partInBoardList.add(layoutPart)

    # mark the space used by this part as taken and adjust the tree of available space
    insertPartInTree(leafNode,layoutPart)
    puts "part added!" if $verbose
    return true
  end
  
  def insertPartInTree(leaf,layoutPart)
    # insert this part into the node of our parent
    leaf.parent.addPart( {:xy => leaf.rect[:xy], :length => layoutPart.getLengthPx, :width => layoutPart.getWidthPx } )
    
    
    #remove the branch not chosen from the parent and link in a new Node on the branch chosen
    # figure out if we are the right or left node and delete the other one - we dont need it anymore cause it wasn't chosen
    if leaf.parent != @root
      if leaf.parent.parent.left == leaf.parent
        leaf.parent.parent.right = nil
      else
        leaf.parent.parent.left = nil
      end
    # else tree is in the degenerative form still - nothing to do
    end
    
    #create a new Parent for the left and right child and its leaves and link it in to our parent. 
    newParent = Node.new(leaf.parent)
    # figure out which side to link this to - depends on where the leaf is 
    if leaf.parent.right == leaf
      leaf.parent.right = newParent
    else
      leaf.parent.left = newParent
    end
    
    # now create a new child to hold the left spaces - link it to the left of our new Parent
    newLeftChild = Node.new(newParent)
    newParent.left = newLeftChild
    
    # and one for the right
    newRightChild = Node.new(newParent)
    newParent.right = newRightChild

    # determine the new origin of the space remaining on the left lower horizontal space
    # left node is the horizontal spaces remaining
    # if width remaining is 0, then we don't need a left node
    newWidth = leaf.rect[:width] - layoutPart.getWidthPx
    if newWidth != 0
      # x remains the same since the part is placed in top left corner
      # y is old y + layout part width + kerf size
      newx = leaf.rect[:xy][0]
      newy = leaf.rect[:xy][1]+layoutPart.getWidthPx
      newLength = leaf.rect[:length]
      # create and add a new left node to the node just created
      newOrigin = Array[newx,newy]
      newLeftChild.addLeft({:xy => newOrigin, :length => newLength, :width => newWidth})
    end
    
    # right node is the other horizontal space remaining to the right of the part placed - the upper horizontal space
    newLength = leaf.rect[:length] - layoutPart.getLengthPx
    if newLength != 0
      # y remains the same since the part is placed in the top left corner - this space is immediately to the right of it
      # x is old x + layout part length
      newy = leaf.rect[:xy][1]
      newx = leaf.rect[:xy][0]+layoutPart.getLengthPx
      newWidth = layoutPart.getWidthPx
      # create and add a new left node to the node just created
      newOrigin = Array[newx,newy]
      newLeftChild.addRight({:xy => newOrigin, :length => newLength, :width => newWidth})
    end
    
    # left node of the right child is the left vertical space below the part placed 
    #right node is the vertical spaces remaining
    # if width remaining is 0, then we don't need a left node
    newWidth = leaf.rect[:width] - layoutPart.getWidthPx
    if newWidth != 0
      # x remains the same since the part is placed in top left corner
      # y is old y + layout part width
      newx = leaf.rect[:xy][0]
      newy = leaf.rect[:xy][1]+layoutPart.getWidthPx
      newLength = layoutPart.getLengthPx
      # create and add a new left node to the node just created
      newOrigin = Array[newx,newy]
      newRightChild.addLeft({:xy => newOrigin, :length => newLength, :width => newWidth})
    end
    
    # right node of the right child is the right vertical space to the right of the part placed 
    newLength = leaf.rect[:length] - layoutPart.getLengthPx
    if newLength != 0
      # y remains the same since the part is placed in top left corner
      # x is old x + layout part length
      newy = leaf.rect[:xy][1]
      newx = leaf.rect[:xy][0]+layoutPart.getLengthPx
      newWidth = leaf.rect[:width]
      # create and add a new left node to the node just created
      newOrigin = Array[newx,newy]
      newRightChild.addRight({:xy => newOrigin, :length => newLength, :width => newWidth})
    end
  end
  
  # use a recursive search to find the best spot for a part
  # best fit means least wastage ie: the area of the piece to place comes closes to the area of the space
  #from the given Node, return a result if it is a leaf node, othewise go left, then go right recursively.
  # returns the chosen best leaf node
  def findBestFit(node,length,width)
    # nothing here if this is a nil node
    return nil if node == nil
    # am i in a node or a leaf
    leaf = false    
    leaf = true if node.getNodeType == "leaf"
    #look at the leaf if we are a leaf
    if leaf
      puts "look at a leaf" if $verbose
      puts "ll=" + node.rect[:length].to_s + "lw="  + node.rect[:width].to_s if $verbose
      puts "pl=" + length.to_s + "pw=" + width.to_s if $verbose
      # won't fit if length is too short
      return nil if length > node.rect[:length]
      # won't fit if width is too narrow
      return nil if width > node.rect[:width]
      # fits, so return as a potential best fit
      return node
    end
    # it's a node
    puts "it's a node" if $verbose
    #go left
    bestLeft = findBestFit(node.left,length,width) 
    #go right
    bestRight = findBestFit(node.right,length,width) 
    # now determine which is the best. No brainer if one or both are nil
    return nil if bestLeft == nil && bestRight == nil
    return bestLeft if bestRight== nil
    return bestRight if bestLeft == nil
    #otherwise both are potentials and we have to choose one
    # pick based on best fit  ie: least amount of waste left or the one further to the origin of the board or both. User can select
    # the bias, otherwise we use both rules and try to pack the parts from left to right
    # ( note area on the board must >= the part size because of earlier decisions )
    area = length*width
    ruleA = ( ((bestLeft.rect[:length]*bestLeft.rect[:width]) - area) <= ((bestRight.rect[:length]*bestRight.rect[:width]) - area)  ) \
            && @layoutOptions[:layoutRuleA]
    ruleB = ( bestLeft.rect[:xy][0] < bestRight.rect[:xy][0] ) && @layoutOptions[:layoutRuleB]
    return bestLeft if ruleA || ruleB
    return bestRight
  end
  
  def getAreaOfParts(node)
    return 0 if node == nil
    return 0 if node.getNodeType == "leaf"
    # add up all parts to the left of the node
    areaLeft = getAreaOfParts(node.left)
    # add up all parts to the right of the node
    areaRight = getAreaOfParts(node.right)
    #puts (areaLeft + areaRight + node.getPartArea).to_s if $verbose1
    return (areaLeft + areaRight + node.getPartArea)
  end
  
  # alternate method of getting the area covered by the parts on a board
  def areaOfPartsOnBoardFromList
    partInBoardListArray = @partInBoardList.getList
    totalArea = 0
    partInBoardListArray.each { |part|
        totalArea += part.getLengthPx*part.getWidthPx
    }
    return totalArea
  end
  
  # Note: This does not seem to return accurate results, though not sure why at this point. Who can find the bug?
  def areaOfPartsOnBoard
    # find all nodes with parts and add up the areas
    return getAreaOfParts(@root)
  end
  
  #return the layout parts on this board
  def getBoardPartList
    return @partInBoardList
  end
  
  def boardFull
    @boardFull
  end
  def setBoardFull
    @boardFull = true
    # when board is declared full, then let's calculate the efficiency
    # add up the area of all of the parts on this board - assume any extra thickness is waste
    # divide by the area of the board
    # we can work these numbers in pixels - good as any - as long as the units are consistent
    #usedArea = areaOfPartsOnBoard
    usedArea = areaOfPartsOnBoardFromList
    @usedAreaPercentage = ((usedArea/@board.getAreaPx)*100).round_to(2)
  end
  def getUsedAreaPercentage
    @usedAreaPercentage
  end
  def board
    @board
  end
  
  def to_s
    return @board.to_s + (' (' + getUsedAreaPercentage.to_s + '%)' if getUsedAreaPercentage != 0 )
  end
  
end


#-----------------------------------------------------------------------------
# class LayoutEngine is the base class for all variants of layout algorithms
#-----------------------------------------------------------------------------
class LayoutEngine  
  # add mixin module to answer any questions we may have which will help us place
  # the parts
  include PartsIntelligence
  
  def initialize(inPartList, inBoardList, boardOptions,layoutOptions,metricVolume)
    # create the perfect store - has only the parts we need :)
    @homerDepot = BoardSuperMarket.new(inBoardList, boardOptions,metricVolume)
    @inPartList = inPartList
    # sort this list
    @inPartList.sort
    #make a copy in case we need to do a do-over
    @originalInPartList = inPartList.deep_clone
    @inBoardList = inBoardList
    @metricVolume = metricVolume
    # create array of boards used to create the cutlist parts assigned to them
    @layoutBoards = Array.new
    @status = LayoutStatus.new("","success")
    @partsToDoList = CutListPartList.new()
    @partsIgnoreList = CutListPartList.new()
    @convergenceImpossible = false
    @layoutOptions = layoutOptions
    @quarters = boardOptions[2]
    @thicknessOptions = (@quarters.include? true )
    #puts "board options=" + @thicknessOptions.to_s
  end

  def convergenceImpossible?
     return @convergenceImpossible == true
  end
    
  def getBoard
    # determine the material and thickness we should be using for this placement
    #if @thicknessOptions
      thickness = @inPartList.getList[0].getThickness 
    #else
      # default thickness
      #thickness=3/4.inch
    #end
    
    # get a board with these properties
    if @layoutOptions[:layoutByMaterial]
      material = @inPartList.getList[0].getMaterial
    else
      material = "generic"
    end
    # get a board from the warehouse
    @board = @homerDepot.getBoard(material, thickness)
    puts @board.to_s if $verbose
    return @board
  end
  
  def placeParts 
    puts "placeParts" if $verbose
    @remainingPartsToPlace = @inPartList.count
    puts "parts to place=" + @remainingPartsToPlace.to_s if $verbose1 || $verbose
    #now get a board with the options specified
    board = getBoard()
    
    #board = getBoard()
    if board != nil
      # make a new layout board using the baord as the basis
      layoutBoard = LayoutBoard.new(board,@layoutOptions,@metricVolume)
      @layoutBoards.push(layoutBoard)
      # remove a part from the list and place it. Stop when no more parts
      while !@inPartList.empty?
        #just remove the parts one at a time for now
        nextPart = @inPartList.removeFirst!
        # place the part on the board
        if ( layoutBoard.addPartToBoard(nextPart) == false )
          #put the part on a todo list?
          puts "part did not fit" if $verbose
          @partsToDoList.add(nextPart)
        end
      end
      
      # now switch the todo list and the inPartsList for the next go
      tempListPtr = @inPartList
      @inPartList = @partsToDoList
      @partsToDoList = tempListPtr
      
      #sort the list again
      @inPartList.sort
      
      # check to see if we are not in a loop because of non-convergence
      # non-convergence here means that a new board has not resulted in any more placed parts -meaning that there is nothing more
      # we can do, so we must stop
      puts "parts left to place=" +  @inPartList.count.to_s if $verbose
      if @remainingPartsToPlace == @inPartList.count
        # we have not made any progress with a new board - must end this foolishness
        @status.set("Some parts were left which could not be placed on given stock.\nTry larger sizes.","warning")
        @convergenceImpossible = true
        # Nothing was added to this board so remove it from our list
        @layoutBoards.pop()
        return
      else
        # this board is full because we've tried every part and no more parts fit but we know we added some
        layoutBoard.setBoardFull
      end      
    end
  end
        
  def run
    
    # keep placing parts until we run out of available boards or there are no more parts
    # to place or there is no more convergence ie: we can't place any more parts regardless of having more boards
    while (@homerDepot.boardsAvailable? && !@inPartList.empty? && !convergenceImpossible?)
      placeParts
      return nil if @status.statusNotGood
      puts  "boards?=" + @homerDepot.boardsAvailable?.to_s + \
            " moreParts?=" + (!@inPartList.empty?).to_s + \
            " converging?=" + (!@convergenceImpossible).to_s if $verbose1 || $verbose
    end
          
    #if we still have parts to place but no boards left then tell the user
    if (!@homerDepot.boardsAvailable? && !@inPartList.empty?)
      @status.set("Some parts were left which could not be placed.\nNo boards left or board limit reached.","warning")
    end
      
    return nil if @status.statusNotGood
        
    # return the result of all our work, a list of layoutBoards, each of which contains the parts
    # to be cut out of them
    return nil, @partsToDoList if @layoutBoards.empty?
    return @layoutBoards, @partsToDoList
  end
end

# There's no real difference betwen the layout engines for solid parts vs sheet goods
# the inBoardList and boardOptions may be different and obviously the parts to place
# are different. Our flavour of layout engine is based mainly on the algorithm used.
class BestFitLayoutEngine < LayoutEngine
end


#-----------------------------------------------------------------------------
# base class for all output generator types, html, or file
# This class is used to drive the different output parts ie: determine what needs to be 
# output and use the appropriate renderers to produce the output
#-----------------------------------------------------------------------------
class OutputDriver
  def initialize(compact,showComps, showSheet, showParts, volumeInMetric,solidParts,sheetParts,hardwareParts,mname)
    @compact = compact
    @showComps = showComps
    @showSheet = showSheet
    @showParts = showParts
    @showVolumeInMetric = volumeInMetric
    @solidParts = solidParts
    @sheetParts = sheetParts
    @hardwareParts= hardwareParts
    @modelName = mname
  end
  
  # open any files required, select a renderer
  def openFiles
    @renderer = nil
  end
  
  def openRenderer
  end
  
  # perform necessary rendering to produce requested output format
  def openParts
    return if (@renderer == nil)
    
    if(@showComps)
      if (@compact)
        @component = CompactComponent.new(@renderer,@showVolumeInMetric)
        @solidParts.sort
      else
        @component = Component.new(@renderer,@showVolumeInMetric)
      end
    end # showComps
    
    if (@showSheet)
      if (@compact)
        @sheet = CompactSheet.new(@renderer,@showVolumeInMetric)
        @sheetParts.sort
      else
        @sheet = Sheet.new(@renderer,@showVolumeInMetric)
      end
    end # showSheet
    
    if @showParts
      @part = CompactPart.new(@renderer,@showVolumeInMetric)
    end
    
  end # openParts
  
  # steps required to produce the requested output
  def run
    openFiles
    openRenderer
    openParts
    render
    close
  end
  
  # close any files which are open
  def closeFiles
  end
  
  # display results of the rendering, either 'output done' msg or display the file
  def displayResults
  end
  
  # steps after rendering is complete
  def close
    closeFiles
    displayResults
  end
end

#-----------------------------------------------------------------------------
# output driver for all file based output
#-----------------------------------------------------------------------------
class FileOutputDriver < OutputDriver
  def initialize(compact,showComps, showSheet, showParts, volumeInMetric,solidParts,sheetParts,hardwareParts,mname,mpath,filename)
    super(compact,showComps, showSheet, showParts, volumeInMetric,solidParts,sheetParts,hardwareParts,mname)
    @mpath = mpath
    @filename = filename
  end
  
  def openFiles
    @namecsv = @mpath + "/" + @modelName + "_" + @filename
    @file = File.new(@namecsv,"w")
  end
  
  def openRenderer
    @renderer = FileRenderer.new()
  end
    
  def render
    if(@showComps && !@solidParts.empty?)
      @file.puts(@component.to_s(@solidParts.getList))
    end
    if(@showSheet && !@sheetParts.empty?)
      @file.puts(@sheet.to_s(@sheetParts.getList))
    end
    if(@showParts && !@hardwareParts.empty?)
      @file.puts(@part.to_s(@hardwareParts))
    end
  end
  
  def closeFiles
    @file.close
  end
  
  def displayResults
    UI.messagebox("Cut List written into: \n\n" + @namecsv + "  \n")
  end

end

#-----------------------------------------------------------------------------
# output driver for cutlist plus output
#-----------------------------------------------------------------------------
class ClpFileOutputDriver < FileOutputDriver
  # perform necessary rendering to produce requested output format
  def openParts
    return if (@renderer == nil)
    
    if(@showComps)
      if (@compact)
        @component = CompactClpComponent.new(@renderer,@showVolumeInMetric)
        @solidParts.sort
      else
        @component = ClpComponent.new(@renderer,@showVolumeInMetric)
      end
    end # showComps
    
    if (@showSheet)
      if (@compact)
        @sheet = CompactClpSheet.new(@renderer,@showVolumeInMetric)
        @sheetParts.sort
      else
        @sheet = ClpSheet.new(@renderer,@showVolumeInMetric)
      end
    end # showSheet
    
    #if @showParts
    #  @part = CompactPart.new(@renderer,@showVolumeInMetric)
    #end
  end
  
  def displayResults
    UI.messagebox("Cut List Plus import file written into: \n\n" + @namecsv + "  \n")
  end
end

#-----------------------------------------------------------------------------
# output driver for all html based output
#-----------------------------------------------------------------------------
class HtmlOutputDriver <OutputDriver
  def openFiles
    @html = ""
  end
  
  def openRenderer
    @renderer = HtmlRenderer.new()
  end
  
  def render
    if(@showComps && !@solidParts.empty?)
      @html = @html+@component.to_s(@solidParts.getList)
    end
    if(@showSheet && !@sheetParts.empty?)
      @html = @html+@sheet.to_s(@sheetParts.getList)
    end
    if(@showParts && !@hardwareParts.empty?)
      @html = @html+@part.to_s(@hardwareParts)
    end
  end
  
  # an opportunity to do something whenever y coordinate reaches a certain value
  # default is to only increment the y coordinate
  def pageYIncrement(yIncrement)
    @y += yIncrement
  end
  
  # no files to close
  def closeFiles
  end
  
  def displayResults
    @resultGui = ResultGui.new
    @resultGui.show(@html)
  end
end

#-----------------------------------------------------------------------------
# output driver for the layout output
#-----------------------------------------------------------------------------
class HtmlLayoutDriver < HtmlOutputDriver
  def initialize(layoutBoards,layoutSheets,unPlacedParts, mname)
    @layoutBoards = layoutBoards
    @layoutSheets = layoutSheets
    @unPlacedParts = unPlacedParts
    @modelName = mname
    @totalEfficiency = 0
  end
  
  def openRenderer
    @renderer = HtmlLayoutRenderer.new(@modelName)
  end
  
  # overwrite the base method as we don't need to open the same parts
  # later we may open the layout Parts list here or something
  def openParts
  end

  def renderLayoutBoards(layoutBoards,type)
      
    if ( layoutBoards != nil && layoutBoards.length > 0 )
      @html += @renderer.sectionHeading(@x,@y,type)
      
      # display board efficiency
      calculateUtilisation(layoutBoards)
      @html += @renderer.displayEfficiency(getTotalEfficiencyString.to_html)
      
      # display board feet required for layout ( # of boards * bdft for each board)
      calculateTotalLayoutBoardFeet(layoutBoards)
      @html += @renderer.displayBoardFeet(getTotalBoardFeetString(layoutBoards).to_html)
      
      # increment y position before starting on the boards
      pageYIncrement(50)
      
      # draw all boards/sheets and layout parts on the boards/sheets
      layoutBoards.each {|layoutBoard| renderBoard(layoutBoard) }
    end
  end
    
  def startPage
    @x = 30
    @y = 30
    @html += @renderer.header(@x,@y)
    pageHeading = "Project: " +@modelName.to_s
    @html += @renderer.pageHeading(@x,@y, pageHeading)
    pageYIncrement(30)
  end
  
  def endPage
    @html += @renderer.footer
  end
  
  def render
    startPage
    
    # if both were nil, then at least say something!
    if ( (@layoutBoards == nil || @layoutBoards.length == 0) && 
         (@layoutSheets == nil || @layoutSheets.length == 0) )
      @html += @renderer.displayText(@x,@y,"No parts to layout")
      pageYIncrement(30)
    else
      renderLayoutBoards(@layoutBoards,"Boards")
      renderLayoutBoards(@layoutSheets,"Sheets") 
    end
    
    #display any uplaced parts
    if @unPlacedParts != nil  && !@unPlacedParts.empty?
      puts "unplaced parts list has " + @unPlacedParts.getList.length.to_s + " entries"
      showUnPlacedParts(@unPlacedParts)
    end
    endPage
  end
  
  def showUnPlacedParts(unPlacedParts)
    # update pageEmpty as we have some unplaced parts to display
    @pageEmpty = false
    unPlacedPartsList = unPlacedParts.getList
    @html += @renderer.displayText(@x,@y,"Parts not placed ( Part(L,W,T) Material):")
    pageYIncrement(30)
    unPlacedPartsList.each { |part|
      @html += @renderer.displayText(@x,@y,part.summary.to_html)
      pageYIncrement(20)
    }
  end

  def calculateTotalLayoutBoardFeet(layoutBoards)
    if layoutBoards != nil
      @layoutBoardFeet = 0
      layoutBoards.each { |layoutBoard|
        @layoutBoardFeet += layoutBoard.board.getBoardFeet
      }
    end
  end

  def getTotalBoardFeetString( layoutBoards )
    if ( layoutBoards != nil && layoutBoards.length > 0 )
      return "Total used for layout: " + @layoutBoardFeet.to_s + " " + ( layoutBoards[0].board.getBoardFeetUnitsString )
    else
      return ""
    end
  end
  
  def calculateUtilisation(layoutBoards)
    # go through each board and get the efficiency of each, average over the number of boards
    if layoutBoards != nil
      x = 0
      efficiency = 0
      @totalEfficiency = 0
      layoutBoards.each { |layoutBoard|
        x += 1
        efficiency += layoutBoard.getUsedAreaPercentage
        @totalEfficiency = (efficiency/x).round_to(2)
      }
    end
  end
  
  def getTotalEfficiencyString
    return "Efficiency = " + @totalEfficiency.to_s + "%"
  end
  
  def renderBoard(layoutBoard)
    # update pageEmpty as soon as we know we have a board to display
    @pageEmpty = false
    #first extract the board dimensions and details and render the board
    board = layoutBoard.board
    @html += @renderer.drawBoard(2,"#cacaaf",layoutBoard.to_s.to_html,@x,@y,board.getLengthPx,board.getWidthPx)
    # now print off each part on this board
    partsList = layoutBoard.getBoardPartList
    partsListArray  = partsList.getList
    partsListArray.each{ |part| renderPart(part) }
    # draw each board 1 at a time - faster output it seems, because it has less to draw each time?
    @html += @renderer.draw
    # move next y coordinate down by board width + 50
    pageYIncrement(board.getWidthPx+50)
  end
  
  def renderPart(part)
    # get the relative coordinates for this part ( returns an array containing x,y)
    coords = part.getLocationOnBoard
    @html += @renderer.drawPart(2,"#bcbc9a",part.getName.to_html,@x+coords[0],@y+coords[1],part.getLengthPx,part.getWidthPx)
  end
  
  # use the layout gui to display the output
  def displayResults
    @resultGui = LayoutGui.new
    #DEBUG
    #puts @html
    #DEBUG
    @resultGui.show(@html)
  end
end

#-----------------------------------------------------------------------------
# SVG output driver for the layout output
#-----------------------------------------------------------------------------
class SvgLayoutDriver < HtmlLayoutDriver
  def initialize(layoutBoards,layoutSheets,unPlacedParts,mname,mpath)
    super(layoutBoards,layoutSheets,unPlacedParts,mname)
    @mpath = mpath
    @filename = "layout"
    @filenameSuffix = ".svg"
    @pageNumber = 1
    @svgFileNameList  = Array.new()
  end
  
  def openRenderer
    @renderer = SvgLayoutRenderer.new(@modelName)
  end
  
  def openFiles
    @html = ""
    @nameSvg = @mpath + "/" + @modelName + "_" + @filename + "-p" + @pageNumber.to_s + @filenameSuffix
    @file = File.new(@nameSvg,"w")
    @pageEmpty = true
  end
  
  def pageYIncrement(yIncrement)
    if ( @y + yIncrement) >= 480
      endPage
      #close this file
      closeFiles
      #increment page number
      @pageNumber +=1
      #open a new file
      openFiles
      startPage
    else
      @y += yIncrement
    end
  end
    
  def displayResults
    svgFileListDisplay = ""
    @svgFileNameList.each { |file| svgFileListDisplay += file }
    fileString = " file:"
    fileString = " files:" if @svgFileNameList.length > 1
    UI.messagebox("SVG Layout written into " + @svgFileNameList.length.to_s + fileString + " \n\n" + svgFileListDisplay + "  \n")
  end
  
  def closeFiles
    @file.puts @html
    @file.close
    if (@pageEmpty == true )
      #delete the file as it is empty
      File.delete(@nameSvg)
    else
      # add this file name to the list of svg files
      @svgFileNameList.push(@nameSvg +"\n")
    end
  end
end

#-----------------------------------------------------------------------------
# Class PartList - holds all selected components which are neither a 
# solid part nor a sheet part - typically this would be hardware...or bryce
#-----------------------------------------------------------------------------
class PartList

  ### Constructor
  def initialize()
    @parts = Array.new
    @partCount = Array.new
  end  ### Constructor

  def parts
    @parts
  end ###def parts

  def partCount
    @partCount
  end ###def partCount

  ### Adds a part to the list.
  def add(inPart)
    index = @parts.index(inPart)
    if(index != nil)
      @partCount[index] = @partCount[index] + 1
    else
      @parts.push(inPart)
      @partCount.push(1)
    end  ###if
  end ###def add
  
  def empty?
    @parts.length == 0
  end
  
  def getList
    @parts
  end

end ###Class PartList

#-----------------------------------------------------------------------------
# cutlist  base class - represent all cutlist parts of the model, be they solid or sheet
# All lengths are stored in sketchup's default inches unless the name says otherwise
#-----------------------------------------------------------------------------
class CutListPart
  def initialize(c, name, material,metricVolume)
    # always get the bounding box from the definition
    # note that we have extended the Sketchup Group class
    # so that we can get the bounding box the same way 
    # whether the entity is a component or a group
    boundingBox = c.bounds
    if c.respond_to? "definition"
     boundingBox = c.definition.bounds
    end
    # get the transformation of the component
    trans=c.transformation.to_a
    scalex=Math.sqrt(trans[0]**2+trans[1]**2+trans[2]**2)
    scaley=Math.sqrt(trans[4]**2+trans[5]**2+trans[6]**2)
    scalez=Math.sqrt(trans[8]**2+trans[9]**2+trans[10]**2)    
    #puts "name=" + name.to_s + "scalex=" + scalex.to_s + " scaley=" + scaley.to_s + " scalez=" + scalez.to_s
    
    width = boundingBox.width*scalex
    height = boundingBox.height*scaley
    depth = boundingBox.depth*scalez
    
    #puts "width=" + width.to_s + " height=" + height.to_s + " depth=" + depth.to_s
    
    #sizes = getSortedArray([boundingBox.width,boundingBox.height,boundingBox.depth])
    sizes = getSortedArray([width,height,depth])
    
    #assume the longest dimension is the length and shortest is the thickness
    @length = sizes[2].inch
    @lengthInFeet = @length.to_feet
    @width = sizes[1].inch
    @thickness = sizes[0].inch
    dimCalculations()
    @material = material
    @name = strip(name,@length.to_s, @width.to_s, @thickness.to_s )
    @canRotate = true
    @metricVolume = metricVolume
    @metric = metricModel?
    @locationOnBoard = nil
  end
  
  def dimCalculations
    @area = @length*@width
    @volume = @area*@thickness
    @squareFeet = @area/144
    @boardFeet = @volume/144
    # part in pixels scale  12in=100px
    @length_px = ((@length/12)*100).to_f.round_to(0)
    @width_px = ((@width/12)*100).to_f.round_to(0)
  end
  
  # Bubble sort
  # sorts in ascending order
  def getSortedArray(array)
    size = array.size()
    pass = size
    for i in (0..pass-2)
      for j in (0..pass-2)
        if (array[j+1] < array[j])
          tmp = array[j]
          array[j] = array[j+1]
          array[j+1] = tmp
        end
      end
    end
    return array
  end

  # used to make the name unique if no name was
  # given to it. The new name is based on the "noname_" + the 
  #concatenation of the dimensions of the part. This allows us
  # to identify identical parts even if they have no name.
  # The order of the dimensions is arbitrary as long as
  # it is consistent but typically same order as the output
  # ie: length, width, thickness
  def strip(name,inV1,inV2,inV3)
    val=name
    if(name=="noname")
      val=("noname_"+inV1+inV2+inV3)
      val = val.gsub(/[ ]/, "_")
      val = val.gsub(/['~"]/, "")
      val = val.gsub(/[\/]/, "-")
    end ## end if
    return val
  end  ## end strip


  #display length or total length of all similar parts
  # is controlled by the 'metricVolume' attribute
  # aka c[6]
  def getTotalLength
    if @metricVolume
      # 1 ft = 0.3048 metres
      (@lengthInFeet*0.3048).round_to(4)
    else
      @lengthInFeet.round_to(2)
    end
  end
  
  # Note: to_l converts inches to Length class, which then prints
  # in whatever units the model is set to
  
  #aka c[1] part length in model units
  def getLengthString
      @length.to_l.to_s
  end
  
  #aka c[2]
  def getWidthString
      @width.to_l.to_s
  end
    
  def getWidth
      @width
  end
    
  def getLength
    @length
  end
    
  #aka c[3]
  def getThicknessString
      @thickness.to_l.to_s
  end

  def getThickness
    @thickness
  end

  #aka c[5] 
  def getMaterial
    @material
  end
  
  #aka c[0]
  def getName
    @name
  end
  
  #aka c[4] for solid parts 
  def getBoardFeet
    if @metricVolume
      #1 mm = 0.0393700787 inches
      # 1 in = 25.40000002590800002642610026955 mm
      # 1bd ft = 144 cu in
      # 1bd ft = 2359.7372232207958956904236222001 cu cm
      # divide by 1000000 to get cu.m.
      #(@boardFeet*2359.7424).round_to(6)
      #(@width.to_l*@length.to_l*@thickness.to_l*2359.7383).round_to(6)
      (@boardFeet*(2359.7372232207958956904236222001/1000000)).round_to(6)
    else
      @boardFeet.round_to(2)
    end
  end
  
  def getBoardFeetLabel
    if @metricVolume
      return "Cubic m"
    else
      return "Board Feet"
    end
  end
  
  # aka c[4] for sheet parts
  def getSquareFeet
    if @metricVolume
      # 1sq ft = 929.0304sq cm
      # divide by 10000 to get sq m
      (@squareFeet*(0.092903040189522201889968968842358)).round_to(6)
    else
      @squareFeet.round_to(2)
    end
  end
  
  def getCanRotate
    @canRotate
  end
  
  def getLengthPx
    @length_px
  end
  
  def getWidthPx
    @width_px
  end
  def addLocationOnBoard(coords)
    @locationOnBoard = coords
  end
  def getLocationOnBoard
    @locationOnBoard
  end
  def changeWidth(width)
    @width = width.inch
    dimCalculations()
  end
  def changeThickness(thickness)
    @thickness = thickness.inch
    dimCalculations()
  end
  
  def convertMeasureForCLP(measure)
    # CutList Plus has some import limitations. Inches, feet, mm and cm are
    # accepted but not meters. So here we convert the measure back to a float in inches
    # if the model is in meters. This is just for CutListPlus import file generation
    if modelInMeters?
      measure = measure.to_f.round_to(4)
    end
    return measure      
  end

  def deep_clone  
   Marshal::load(Marshal.dump(self))  
  end 
 
  def summary
    return (getName + " (" + getLengthString + ", " + getWidthString + ", " + getThicknessString + ") " + getMaterial)
  end
end

#-----------------------------------------------------------------------------
# Class SolidPart - to represent all solid parts which are part of the cut list and layout
# Derived from base class CutListPart
#-----------------------------------------------------------------------------
class SolidPart < CutListPart
  
  # initialization for class SolidPart
  # c - sketchup entity, either component or group
  # name - is a string of the name of the entity
  # material - is a string of the material for this entity
  # nominalMargin is a number in 16ths of the allowance required in the thickness over the final part size
  # quarter is an array of 4 elements, being boolean values of fourq, fiveq, sixq and eightq respectively as entered by the user
  def initialize(c, name, material, nominalMargin, quarter,nominalOut,metricVolume)
    @nominalMargin = nominalMargin.inch
    @nominalOut = nominalOut
    @quarter = quarter
    super(c,name,material,metricVolume)
  end
  
  def dimCalculations
    super
    @nominalThickness = handleNominalSize(@thickness).inch
    @volume = @area*@nominalThickness
    @boardFeet = @volume/144
  end
  
    ## Turn a thickness into a nominal thickness ##
  def handleNominalSize(inThickness)
    result = inThickness
    #keep marginThickness for use later in the layout
    @marginThickness = @nominalMargin+inThickness
    @marginThickness = @marginThickness.round_to(4)
    if((@quarter[0]) && @marginThickness<=1)
      result = 1.inch
    elsif((@quarter[1]) && @marginThickness<=1.25)
      result = (1.25).inch
    elsif((@quarter[2])&& @marginThickness<=1.5)
      result = (1.5).inch
    elsif((@quarter[3]) && @marginThickness<=2)
      result = 2.inch
    elsif((@quarter[4]) && @marginThickness<=2.5)
      result = (2.5).inch
    end #if
    return result
  end ## handleNominalSize
  
  def getThicknessString
    if @nominalOut
      @nominalThickness.to_l.to_s
    else
      @thickness.to_l.to_s
    end
  end
  
  def getMarginThickness
    @marginThickness
  end
  
  def getThickness
    @nominalThickness
  end
  
end

#-----------------------------------------------------------------------------
# Class SheetPart - to represent all components cut from sheet stock 
# and which are part of the cut list and layout. Derived from base class CutListPart
#-----------------------------------------------------------------------------
class SheetPart < CutListPart
  
  # for sheet goods, the board ft measure is actually the square footage
  def getBoardFeet
    getSquareFeet
  end
  
  def getBoardFeetLabel
    if @metricVolume
      return "Square m"
    else
      return "Square feet"
    end
  end

end

#-----------------------------------------------------------------------------
# Base class for all lists of components 
# Maintains a list of all parts which can be sorted by name within length within board feet
# simultaneously maintains a list of parts indexed by [material,thickness]
#-----------------------------------------------------------------------------
class CutListPartList
  # Constructor
  def initialize()
    # array of cut list part objects
    @componentList = Array.new
    @componentListByMaterialAndThickness = Hash.new
  end 
  
  def addToPartDatabase(cutListPart)
    return if cutListPart == nil
    material = cutListPart.getMaterial
    # get thickness and convert to string to use as an index in the hash
    thickness = cutListPart.getThickness.inch.to_s
    #puts " thickness=" +  thickness.to_s
    if !@componentListByMaterialAndThickness.include? material
      # this a new material, create a new entry in the material array
      newThicknessHash = Hash.new
      #puts "new material hash for " + material.to_s
      @componentListByMaterialAndThickness[material] = newThicknessHash
    end
    # If this is a new thickness, create a new parts array and a new hash to this array
    # using this new thickness
    if !@componentListByMaterialAndThickness[material].include? thickness
      #create a new array for the parts at this thickness
      #puts "new thickness hash for " + thickness.to_s + " " + material.to_s
      newPartArray = Array.new
      @componentListByMaterialAndThickness[material][thickness] = newPartArray
    end
    # in any case, add this part
    @componentListByMaterialAndThickness[material][thickness].push( cutListPart )
  end
  
  def removeFromPartDatabase(cutListPart)
    return if cutListPart == nil
    # when removing we must first update the hash list of [material,thickness]
    material = cutListPart.getMaterial
    # get thickness and convert to string to use as an index in the hash
    thickness = cutListPart.getThickness.inch.to_s
    #puts "removing part from material=" + material + " thickness= " + thickness
    #puts @componentListByMaterialAndThickness[material].to_s
    #puts @componentListByMaterialAndThickness[material][thickness].to_s
    @componentListByMaterialAndThickness[material][thickness].delete( cutListPart )
    #remove the thickness hash if now empty
    if @componentListByMaterialAndThickness[material][thickness].empty?
      #remove the thickness hash
      @componentListByMaterialAndThickness[material].delete(thickness)
    end
    #remove the material hash if now empty
    if @componentListByMaterialAndThickness[material].empty?
      #remove the material hash
      @componentListByMaterialAndThickness.delete(material)
    end
  end
  
  # add a cut list object to the array
  def add(cutListPart)
    @componentList.push(cutListPart)
    addToPartDatabase(cutListPart)
  end
  
  def remove(cutListPart)
    removeFromPartDatabase(cutListPart)
    @componentList.delete(cutListPart)
  end
  
  def empty?
    @componentList.length == 0
  end
  
  def count
    @componentList.length
  end
  
  def allPartsSameMaterial?
    return true if @componentList.empty?
    material = @componentList.first.getMaterial
    return ( @componentList.select{|x| x.getMaterial != material } == nil )
  end
  
  def allPartsSameThickness?
    return true if @componentList.empty?
    thickness = @componentList.first.getThickness
    return (@componentList.select{|x| x.getThickness != thickness } == nil )
  end
  
  # returns the list in its current context
  def getList
      return @componentList
  end
  
  def removeFirst!
    #removes the first part in the list (biggest part in a sorted list) and return it.
    # returns nil if the list is empty
    cutListPart = @componentList.shift
    removeFromPartDatabase(cutListPart)
    return cutListPart
  end
  
  def insertFirst!(cutListPart)
    #add a part back to the front of the list
    @componentList.unshift(cutListPart)
    addToPartDatabase(cutListPart)
  end
    
  # sort and return a copy of the list. Does not change the component list 
  def sort
    # sort in descending order of board feet ( which is sq feet for sheets )
    # if board feet of the two parts are the same, then sort by length
    # if length is the same, then sort by name - putting like parts together
    # return the sorted list
    sortedComponentList = @componentList
    size = sortedComponentList.size()
    pass = size
    for i in (0..pass-2)
      for j in (0..pass-2)
        if (sortedComponentList[j].getBoardFeet < sortedComponentList[j+1].getBoardFeet)
          tmp = sortedComponentList[j+1]
          sortedComponentList[j+1] = sortedComponentList[j]
          sortedComponentList[j] = tmp
        elsif (sortedComponentList[j].getBoardFeet == sortedComponentList[j+1].getBoardFeet)
          # if boardfeet is identical, then sort by length
          if (sortedComponentList[j].getTotalLength < sortedComponentList[j+1].getTotalLength )
            tmp = sortedComponentList[j+1]
            sortedComponentList[j+1] = sortedComponentList[j]
            sortedComponentList[j] = tmp
          elsif (sortedComponentList[j].getTotalLength == sortedComponentList[j+1].getTotalLength )
            # if length identical, then sort by name
            if (sortedComponentList[j].getName > sortedComponentList[j+1].getName )
              tmp = sortedComponentList[j+1]
              sortedComponentList[j+1] = sortedComponentList[j]
              sortedComponentList[j] = tmp
            end
          end
        end
      end
    end
    return sortedComponentList
  end
  
  # sort and overwrite the list with the sorted list
  def sort!
    sortedList = @componentList.sort
    @componentList = sortedList
  end
  
  def deep_clone  
   Marshal::load(Marshal.dump(self))  
  end 
    
  def splitPartsListByMaterial
    @listOfPartsListsByMaterial = Array.new
    # the default is a single entry in the array with the original partsList if all parts have the same material
    if allPartsSameMaterial?
        @listOfPartsListsByMaterial.push(self)
        return @listOfPartsListsByMaterial 
    end
    # some parts are of a different material. Retrieve those. Each new material gets its own partsList
    # all thicknesses are lumped together
    @componentListByMaterialAndThickness.each{ |material,thickness|
      partListByMaterial = CutListPartList.new
      @componentListByMaterialAndThickness[material].each{ |thickness,part|
        @componentListByMaterialAndThickness[material][thickness].each{ |part|
          partListByMaterial.add(part)
        }
      }
      @listOfPartsListsByMaterial.push(partListByMaterial)
    }
    return @listOfPartsListsByMaterial
  end
      
  def splitPartsListByThicknessAndMaterial
    @listOfPartsListsByMaterialAndThickness = Array.new
    # the default is a single entry in the array with the original partsList if all parts have the same material and thickness
    if allPartsSameMaterial? && allPartsSameThickness?
        @listOfPartsListsByMaterialAndThickness.push(self)
        return @listOfPartsListsByMaterialAndThickness
    end
    # else some parts are of a different material. Retrieve those. Each new combination of material,thickness
    # gets its own list
    @componentListByMaterialAndThickness.each{ |material,thickness|
      @componentListByMaterialAndThickness[material].each{ |thickness,part|
        partListByThicknessAndMaterial = CutListPartList.new
        @componentListByMaterialAndThickness[material][thickness].each{ |part|
          partListByThicknessAndMaterial.add(part)
        }
        @listOfPartsListsByMaterialAndThickness.push(partListByThicknessAndMaterial)
      }
    }
    #puts @listOfPartsListsByMaterialAndThickness.to_s
    return @listOfPartsListsByMaterialAndThickness
  end
  
  def splitPartsListByThickness
    @listOfPartsListsByThickness = Array.new
    if allPartsSameThickness?
        @listOfPartsListsByThickness.push(self)
        return @listOfPartsListsByThickness
    end
    # else some parts are of a different thickness. Retrieve those. Each new thickness regardless of material gets its own list
    # search the database and create a hash by thickness, which is easier to populate, then convert to an array at the end.
    thicknessHash = Hash.new
    @componentListByMaterialAndThickness.each{ |material,thickness|
      @componentListByMaterialAndThickness[material].each{ |thickness,part|
        if !thicknessHash.include? thickness
          puts "List Parts by Thickness: new thickness=" +  thickness
          partListByThickness = CutListPartList.new
          thicknessHash[thickness] = partListByThickness
        end
        @componentListByMaterialAndThickness[material][thickness].each{ |part|
          thicknessHash[thickness].add(part)
        }
      }
    }
    #convert hash to our array
    thicknessHash.each { |thickness,partList| @listOfPartsListsByThickness.push(partList) }
    return @listOfPartsListsByThickness
  end 
  
  def +(cutListPartList)
    # add the parts on the list and return self.
    return if cutListPartList == nil
    cutListPartList.getList.each { |part| add(part) }
  end
  
  
end
  
#-----------------------------------------------------------------------------
# Class SolidPartList - List of all solid component parts which are part 
# of the cut list and layout
#-----------------------------------------------------------------------------
class SolidPartList < CutListPartList
end

#-----------------------------------------------------------------------------
# Class SheetPartList - List of all sheet component parts which are part 
# of the cut list and layout
#-----------------------------------------------------------------------------
class SheetPartList < CutListPartList
end

#-----------------------------------------------------------------------------
# Class BoardList - List of all boards available for the layout where 
# boards have been specified by the user. A list of raw boards.
#-----------------------------------------------------------------------------
class BoardList < CutListPartList
end

#-----------------------------------------------------------------------------
# Gui base class to define some common things on all the GUIs
#-----------------------------------------------------------------------------
class GuiBase
  #title used for all html pages
  @@title = "Cut List v4.0.7"
  
  # relative location of the cutlist input html page
  @@cutlistui_location = '/cutlistui/cutlistui.html'
  
  #relative location of the cutlist result html template
  @@cutlistresult_location = '/cutlistui/cutlistresult.html'
  
  #relative location of the configuration save file
  @@cutlistConfig_location = '/cutlistui/config.txt'
  
  def getHtmlTitle
    return @@title
  end
  
  def getUiHtmlLocation
    return @@cutlistui_location 
  end
  
  def getResultHtmlLocation
    return @@cutlistresult_location
  end
  
  def getConfigLocation
    return @@cutlistConfig_location
  end
  
  def show(results)
    @results = results
    openDialog
    addCallbacks
    display
    return nil
  end ##ResultGui.displayResults

end

#-----------------------------------------------------------------------------
# class WebGui - for user to select the options and run the script from an html page 
# This dialog is what is displayed when the user first clicks on the plugin
#-----------------------------------------------------------------------------
class WebGui < GuiBase
  def openDialog
    @dlg = UI::WebDialog.new(getHtmlTitle, true)
    @dlg.set_file( File.dirname(__FILE__)+getUiHtmlLocation)
  end
  
  def addCallbacks
    @dlg.add_action_callback("handleRun") {|d,p|
      # DEBUG - print selected input parameters from dialog
      #puts p
      parse_input_string(p)
      reporter = Reporter.new
      reporter.sketchup6Init(@cutlist_options,@layout_options)
    }
    @dlg.add_action_callback("handleClose") {|d,p|
      @dlg.close()
    }
    @dlg.add_action_callback("handleSaveConfig"){|d,p| 
      parse_input_string(p)
      stdout = $stdout
      File.open(File.dirname(__FILE__)+getConfigLocation,File::CREAT|File::WRONLY) do |f|
        f.puts p.to_s
        f.puts @cutlist_options.inspect
        f.puts @layout_options.inspect
      end
      $stdout = stdout
      puts "Saved to " + File.dirname(__FILE__)+getConfigLocation if $verbose1
    }
  end
  
  def display
    @dlg.show {}
  end
  
  def start
    @results=""
    show(@results)
    return nil
  end # start
  
  def calcNominalMargin(nominalMargin)
    if(nominalMargin=="1")
      return 1.inch/16
    elsif(nominalMargin=="2")
      return 2.inch/16
    elsif(nominalMargin=="3")
      return 3.inch/16
    elsif(nominalMargin=="4")
      return 4.inch/16
    elsif(nominalMargin=="5")
      return 5.inch/16
    else
      return 0
    end ## if
  end

  def calcKerfSize(kerfSize)
    if (kerfSize == "2sixteenth")
      return 2.inch/16
    elsif (kerfSize == "3sixteenth")
      return 3.inch/16
    else
       return 0
    end
  end
  
  def parse_input_string(p)
    puts p.to_s if $verbose
    a = p.split(',')
    parse_input_array(a)
    @cutlist_options = getCutlistOptions(a)
    @layout_options = getLayoutOptions(a)
    puts @layout_options.to_s if $verbose
  end
  
  def parse_input_array(a)
    @compactList = a[0]
    @nominalOut  = a[1]
    @linearFeet  = a[2]
    @outFileUsed = a[3]
    @outFileName = a[4]
    @clpFileUsed = a[5]
    @clpFileName = a[6]
    @printPage   = a[7]
    @partWords   = a[8]
    @sheetWords  = a[9]
    @comps       = a[10]
    @sheet       = a[11]
    @parts       = a[12]
    @forceBoardFeet = a[13]
    @fourq       = a[14]
    @fiveq       = a[15]
    @sixq        = a[16]
    @eightq      = a[17]
    @tenq      = a[18]
    @nominalMargin = a[19]
    @layout          =a[20]
    @boardWidthOption =a[21]
    @boardLengthOption =a[22]
    @sheetWidthOption =a[23]
    @sheetLengthOption =a[24]
    @nominalWidth  = a[25]
    @kerfSize      = a[26]
    @splitWideParts = a[27]
    @splitThickParts = a[28]
    @layoutRuleA     = a[29]
    @layoutRuleB     = a[30]
    @layoutByMaterial = a[31]
    @displayUnplacedParts = a[32]
    @svgLayout = a[33]
  end
  
  def getCutlistOptions(a)            
    cutlist_options = {
      :compactList => (@compactList == "compactList"),
      :listAllSorted => (@compactList == "listAllSorted"),
      :linearFeet => (@linearFeet == "true"),
      :outFileUsed => (@outFileUsed == "true"),
      :clpFileUsed => (@clpFileUsed == "true"),
      :printPage => (@printPage == "true"),
      :showComps => (@comps == "true"),
      :showSheet => (@sheet == "true"),
      :showParts => (@parts == "true"),
      :forceBoardFeet => (@forceBoardFeet == "true"),
      :layout => (@layout == "true"),
      :partWords => @partWords.split(' '),
      :sheetWords => @sheetWords.split(' '),
      :svgLayout =>(@svgLayout == "true")
    }
      
      # include file names only if not nil
      if @outFileName != ""
        cutlist_options[:outFileName] = @outFileName
      end
      
      if @clpFileName != ""
        cutlist_options[:clpFileName] = @clpFileName
      end
      
      return cutlist_options
    end
    
  def getLayoutOptions(a)
    layout_options = {
      :fourq => (@fourq == "true"),
      :fiveq => (@fiveq == "true"),
      :sixq => (@sixq == "true"),
      :eightq => (@eightq == "true"),
      :tenq => (@tenq == "true"),
      :nominalMargin =>  calcNominalMargin(@nominalMargin),
      :nominalOut => (@nominalOut == "true"),
      :board4w=>(@boardWidthOption == "board4w"),
      :board6w=>(@boardWidthOption == "board6w"),
      :board8w=>(@boardWidthOption == "board8w"),
      :board10w=>(@boardWidthOption == "board10w"),
      :board12w=>(@boardWidthOption == "board12w"),
      :board2l=>(@boardLengthOption == "board2l"),
      :board4l=>(@boardLengthOption == "board4l"),
      :board6l=>(@boardLengthOption == "board6l"),
      :board8l=>(@boardLengthOption == "board8l"),
      :board10l=>(@boardLengthOption == "board10l"),
      :board12l=>(@boardLengthOption == "board12l"),
      :sheet2w=>(@sheetWidthOption == "sheet2w"),
      :sheet4w=>(@sheetWidthOption == "sheet4w"),
      :sheet2l=>(@sheetLengthOption == "sheet2l"),
      :sheet4l=>(@sheetLengthOption == "sheet4l"),
      :sheet8l=>(@sheetLengthOption == "sheet8l"),
      :nominalWidth=>(@nominalWidth == "false"),
      :kerfSize=> calcKerfSize(@kerfSize),
      :splitWideParts=>(@splitWideParts == "true"),
      :splitThickParts=>(@splitThickParts == "true"),
      :layoutRuleA=>(@layoutRuleA == "true"),
      :layoutRuleB=>(@layoutRuleB == "true"),
      :layoutByMaterial=>(@layoutByMaterial == "true"),
      :displayUnplacedParts=>(@displayUnplacedParts == "true")
    }
    return layout_options
  end
  
end ## WebGui class

# class ResultGui - for the output of the user selection when html output has been selected
class ResultGui < GuiBase
  
  def openDialog
    @resDialog = UI::WebDialog.new(getHtmlTitle, true)
    @resDialog.set_file( File.dirname(__FILE__)+getResultHtmlLocation)
    @resDialog.set_position(150,150)
  end
  
  def addCallbacks
    @resDialog.add_action_callback("handleClose") {|d,p| @resDialog.close() }
    @resDialog.set_on_close {
      @resDialog.execute_script("handleResults('No results');");
    }
  end
  
  def display
    @resDialog.show {
      @resDialog.execute_script("handleResults(\'#{@results}\');");
    }
  end
  
end ## ResultGui class

class LayoutGui < ResultGui 
  
  def openDialog
    super
    @resDialog.set_position(200,200)
  end
  
  def addCallbacks
    @resDialog.add_action_callback("handleClose") {|d,p| @resDialog.close() }
  end
  
  def display
    @resDialog.show {
      @resDialog.execute_script("handleLayoutScript('"+"#{@results}" +"');");
    }
  end

end


#########################
# Thing superclass      #
#########################
class Thing

  def initialize(inRenderer, inMetric)
    @renderer = inRenderer
    @metric = inMetric
  end ## end initialize

  def getTitle(title)
    return @renderer.getTitle(title)
  end ## end getTitle

  def getHeaderRow(headers)
    return @renderer.getHeaderRow(headers)
  end ## end getHeaderRow

  def getFooterRow()
    return @renderer.getFooterRow()
  end ## end getFooterRow

  def getRow(columns)
    return @renderer.getRow(columns)
  end ## end getRow

  def getAmount(amount)
    return @renderer.getAmount(amount)
  end ## end getArea

  def getBlankLine()
    return @renderer.getBlankLine()
  end ## end getBlankLine

  def to_s(inList)
  end ## end to_s

end ##class Thing

#########################
# Component             #
#########################
class Component < Thing
  
  def initialize(inRenderer, inMetric)
    super(inRenderer, inMetric)
      @measureLabel = getMeasureLabel
      @measureUnits = getMeasureUnits
  end
  
  def getMeasureLabel
    if @metric
      return "Cubic m"
    else
      return "Board Foot"
    end
  end
  
  def getMeasureUnits
    if @metric
      return "m"
    else
      return "Feet"
    end
  end
  
  def getPartPrefix()
    return "C-"
  end

  def getTitleName()
    return "Components";
  end

  def getHeadingArray()
      headings=["Part #","Description","Length(X)","Width(Y)","Thickness(Z)",@measureLabel,"Material"]
   return headings
  end

  def getAmountTitleName()
      return "Total Component " + @measureLabel
  end

  def getMaterialTitleName()
    return "Component Materials"
  end

  def isAmountEnabled()
    return true
  end

  def isMaterialEnabled()
    return true
  end
  
  def processRows(inList)

    component=""
    cols = Array.new
    i = 1
    ii = 0
    ix = ""
    cx = ""
    
    for c in inList
      i = i + 1 if c.getName != cx and cx != ""
      ii = 1 if c.getName != cx
      ix = i.to_s+"-"+ii.to_s
      cols[0]=getPartPrefix()+ix
      cols[1]=c.getName
      cols[2]=c.getLengthString
      cols[3]=c.getWidthString
      cols[4]=c.getThicknessString
      cols[5]=c.getBoardFeet.to_s
      cols[6]=c.getMaterial
      component = component + getRow(cols)
      ### gives sub-part numbers to same named compos; last = total
      ii = ii + 1
      cx = c.getName
      ## Add the board feet
      @totalBF = @totalBF + c.getBoardFeet
      inList = false
      for d in @materialList
        if(d[0]==c.getMaterial)
          d[1] = d[1]+c.getBoardFeet
          inList = true
        end  ## end if
      end  ## end for

      if(!inList)
            @materialList = @materialList.push([c.getMaterial,c.getBoardFeet])
      end
          
    end ## end for
                          
    return component
  end ## end processRows

  def to_s(inList)

    @materialList = []
    @totalBF = 0
    @totalLength = 0
    roundToNumOfDigits = 2
    roundToNumOfDigits = 4 if(@metric)
      
    component = ""
    component = component + getTitle(getTitleName())
    headings = getHeadingArray()
    component = component + getHeaderRow(headings)
    component = component + processRows(inList)
    component = component + getFooterRow()
    component = component + getBlankLine()

    ## Total Board Feet Table
    if(isAmountEnabled())
      component = component + getTitle(getAmountTitleName())
      cols = Array.new
      cols[0] = @totalBF.round_to(roundToNumOfDigits).to_s
      component = component + getRow(cols)
      component = component + getFooterRow()
      component = component + getBlankLine()
    end

    ## Materials Table
    if(isMaterialEnabled())
      component = component + getTitle(getMaterialTitleName())
      headings=["Material",@measureLabel]
      component = component + getHeaderRow(headings)
      for d in @materialList
        cols = Array.new
        cols[0] = d[0]
        cols[1] = d[1].round_to(roundToNumOfDigits).to_s
        component = component + getRow(cols)
      end ## end for
      component = component + getFooterRow()
      component = component + getBlankLine()
    end

    return component

  end ## end to_s

end ##class Component

#########################
# CompactComponent      #
#########################
class CompactComponent < Component

  def getHeadingArray()
    headings=["Part#","Quantity","Description","Length(X)","Width(Y)","Thickness(Z)",@measureLabel + " (per)",
    @measureLabel + " (total)", "Total Length (" + @measureUnits + ")", "Material"]
    return headings
  end

  def processRows(inList)

    component=""
    partId = 1
    partCount = 1
    lastPart = ["",0,0,0]
    cols = Array.new
    row=""

    for c in inList

# If parts match the name and dimensions, then they are considered the
# same and will be displayed in the compact form
      if (  ( c.getName == lastPart[0] )            &&
            ( c.getLengthString == lastPart[1] )   &&
            ( c.getWidthString  == lastPart[2] )   &&
            ( c.getThicknessString == lastPart[3] ) )
        partCount = partCount + 1
      elsif(lastPart != ["",0,0,0])
        component = component + row
        partId = partId + 1
        partCount = 1
      end ##if

      cols[0]=getPartPrefix()+partId.to_s
      cols[1]=partCount.to_s
      cols[2]=c.getName
      cols[3]=c.getLengthString
      cols[4]=c.getWidthString
      cols[5]=c.getThicknessString
      cols[6]=c.getBoardFeet.to_s
      cols[7]=((c.getBoardFeet)*(partCount)).to_s
      cols[8]=((c.getTotalLength)*(partCount)).to_s
      cols[9]=c.getMaterial
      row = getRow(cols)
      lastPart = [c.getName,c.getLengthString,c.getWidthString,c.getThicknessString]

      ## Add the square feet
      @totalBF = @totalBF + c.getBoardFeet
      inList = false
      for d in @materialList
        if(d[0]==c.getMaterial)
          d[1] = d[1]+c.getBoardFeet
          inList = true
        end
      end

      if(!inList)
        @materialList = @materialList.push([c.getMaterial,c.getBoardFeet])
      end

    end#for c

    ##Output last row
    component = component + row

    return component
  end ## end processRows


end ##class CompactComponent

#-----------------------------------------------------------------------------
#########################
# CompactPart           #
#########################
class CompactPart < CompactComponent

  def getPartPrefix()
    return "P-"
  end

  def getTitleName()
    return "Other Parts";
  end

  def getHeadingArray()
   headings=["Part #","Quantity","Description"]
   return headings
  end

  def getAmountTitleName()
    return ""
  end

  def getMaterialTitleName()
    return ""
  end

  def isAmountEnabled()
    return false
  end

  def isMaterialEnabled()
    return false
  end
  
  def processRows(inList)

    component=""
    cols = Array.new

    if(inList.parts.length > 0)
      for e in 0..(inList.parts.length-1)
        cols[0]=getPartPrefix()+(e+1).to_s
        cols[1]=inList.partCount[e].to_s
        cols[2]=inList.parts[e]
        component = component + getRow(cols)
      end  ###for
    end ###if
    return component

  end ## end processRows


end ## CompactPart



#########################
# CompactSheet          #
#########################
class CompactSheet < CompactComponent
  
  def getMeasureLabel
    if @metric
      return "Square m"
    else
      return "Square Foot"
    end
  end
  
  def getPartPrefix()
    return "S-"
  end

  def getHeadingArray()
      headings=["Part#","Quantity","Description","Length(X)","Width(Y)","Thickness(Z)",@measureLabel + " (per)",
      @measureLabel + " (total)", "Total Length (" + @measureUnits +")","Material"]
   return headings
  end

  def getTitleName()
    return "Sheet Goods";
  end

  def getAmountTitleName()
      return "Total Sheet " + @measureLabel
  end

  def getMaterialTitleName()
    return "Sheet Materials"
  end

end ##class CompactSheet


#########################
# ClpComponent          #
#########################
class ClpComponent < Component

  def getTitleName()
    return "";
  end

  def getHeadingArray()
    headings=["Part #","Sub-Assembly", "Description", "Copies",
    "Thickness(Z)","Width(Y)","Length(X)","Material Type", "Material Name", "Can Rotate"]
    return headings
  end

  def getAmountTitleName()
    return ""
  end

  def getMaterialTitleName()
    return ""
  end

  def isAmountEnabled()
    return false
  end

  def isMaterialEnabled()
    return false
  end

  def getMaterialType()
    return "DL"
  end
  
  def processRows(inList)
    component=""
    cols = Array.new
    i = 1
    ii = 0
    ix = ""
    cx = ""
    for c in inList
      i = i + 1 if c.getName != cx and cx != ""
      ii = 1 if c.getName != cx
      ix = i.to_s+"-"+ii.to_s
      cols[0]=getPartPrefix()+ix
      cols[1]=c.getName
      cols[2]=c.getName
      cols[3]="1"
      # CutListPlus accepts all sketchup units except meters
      # If inches, then the inches symbol " needs to be stripped with to_clp since clp won't accept it - this is probably a CLP bug
      # If meters then we will convert to inches - this is sorted with the convertMeasureForCLP method
      cols[4]=c.convertMeasureForCLP(c.getThickness).to_s.to_clp
      cols[5]=c.convertMeasureForCLP(c.getWidth).to_s.to_clp
      cols[6]=c.convertMeasureForCLP(c.getLength).to_s.to_clp
      cols[7]=getMaterialType()
      cols[8]=c.getMaterial
      cols[9]="yes"
      component = component + getRow(cols)
      ### gives sub-part numbers to same named compos; last = total
      ii = ii + 1
      cx = c.getName
    end ## end for

    return component
  end ## end processRows


end ##class ClpComponent

#########################
# ClpSheet              #
#########################
class ClpSheet < ClpComponent

  def getPartPrefix()
    return "S-"
  end

  def getMaterialType()
    return "SG"
  end

end ##class ClpComponent


#########################
# CompactClpComponent   #
#########################
class CompactClpComponent < CompactComponent

  def getTitleName()
    return "";
  end

  def getHeadingArray()
   headings=["Part #","Sub-Assembly", "Description", "Copies",
   "Thickness(Z)","Width(Y)","Length(X)","Material Type", "Material Name", "Can Rotate"]
   return headings
  end

  def getAmountTitleName()
    return ""
  end

  def getMaterialTitleName()
    return ""
  end

  def isAmountEnabled()
    return false
  end

  def isMaterialEnabled()
    return false
  end

  def getMaterialType()
    return "DL"
  end

  def processRows(inList)

    component=""
    partId = 1
    partCount = 1
    lastPart = ""
    cols = Array.new
    row=""

    for c in inList

      if (c.getName == lastPart)
        partCount = partCount + 1
      elsif(lastPart != "")
        component = component + row
        partId = partId + 1
        partCount = 1
      end ##if

      cols[0]=getPartPrefix()+partId.to_s
      cols[1]=c.getName
      cols[2]=c.getName
      cols[3]=partCount.to_s
      
      # CutListPlus accepts all sketchup units except meters
      # If inches, then the inches symbol " needs to be stripped with to_clp since clp won't accept it - this is probably a CLP bug
      # If meters then we will convert to inches - this is sorted with the convertMeasureForCLP method
      cols[4]=c.convertMeasureForCLP(c.getThickness).to_s.to_clp
      cols[5]=c.convertMeasureForCLP(c.getWidth).to_s.to_clp
      cols[6]=c.convertMeasureForCLP(c.getLength).to_s.to_clp
      cols[7]=getMaterialType()
      cols[8]=c.getMaterial
      cols[9]="yes"
      row = getRow(cols)
      lastPart = c.getName
    end#for c

    ##Output last row
    component = component + row
    return component
  end ## end processRows

end ##class CompactClpComponent


#########################
# CompactClpSheet       #
#########################
class CompactClpSheet < CompactClpComponent

  def getPartPrefix()
    return "S-"
  end

  def getMaterialType()
    return "SG"
  end

end ##class CompactClpSheet

#########################
# Sheet                 #
#########################
class Sheet < Component
  
  def getMeasureLabel
    if @metric
      return "Square m"
    else
      return "Square Foot"
    end
  end

  def getPartPrefix()
    return "S-"
  end
  def getTitleName()
    return "Sheet Goods";
  end

  def getHeadingArray()
    headings=["Part #","Description","Length(X)","Width(Y)","Thickness(Z)",@measureLabel, "Material"]
    return headings
  end

  def getAmountTitleName()
    return "Total Sheet " + @measureLabel
  end

  def getMaterialTitleName()
    return "Sheet Materials"
  end

end ##class Sheet


#-----------------------------------------------------------------------------
#########################
# Renderer superclass   #
#########################
class Renderer

  def getTitle(title)
  end ## end getTitle

  def getHeaderRow(headers)
  end ## end getHeaderRow

  def getFooterRow()
  end ## end getFooterRow

  def getRow(columns)
  end ## end getRow

  def getAmount(amount)
  end ## end getAmount

  def getBlankLine()
  end ## end getBlankLine

end ## end class Renderer

#########################
# HtmlRenderer          #
#########################
class HtmlRenderer < Renderer

  def getTitle(title)
    colspan = 2
    html = ""
    html = html+"<BR><B><H4 style=\"color:#7d7d4f\">"+title+"</H4></B>"
   # html = html+"<table cellpadding=2 cellspacing=0 border=1 frame=border >\\n"
   # html = html+"<table cellpadding=2 cellspacing=0 frame=hsides bgcolor=#ECF0F3>\\n"
    html = html+"<table id=\"cutlistTable\" cellpadding=4 cellspacing=0 frame=border border=\"1\" bordercolor=#8c8c8c rules=groups bgcolor=#ECF0F3>"
    return html
  end ## end getTitle

  def getHeaderRow(headers)
    html = ""
    html = html+"<tr>"
    for h in headers
      html = html+"<td>"
      html = html+"<B>"+h+"</B>"
      html = html+"</td>"
    end ## end for
    html = html+"</tr>"
    return html
  end ## end getHeaderRow

  def getFooterRow()
    html = ""
    html = html+"</table>"
    return html
  end ## end getFooterRow

  def getRow(columns)
    html = ""
    html = html+"<tr>"
    for c in columns
      html = html+"<td>"
      html = html+c.to_html
      html = html+"</td>"
    end ## end for
    html = html+"</tr>"
    return html
  end ## end getRow

  def getAmount(amount)
    html = ""
    html = html+"<tr>"
    html = html+"<td>"
    html = html+amount
    html = html+"</td>"
    html = html+"</tr>"
    return html
  end ## end getArea

  def getBlankLine()
   return "<BR>"
  end ## end getBlankLine

end ## end classHtmlRenderer

#-----------------------------------------------------------------------------
# Layout renderer base class
#-----------------------------------------------------------------------------
class LayoutRenderer
  def initialize(modelName)
    @modelName = modelName
  end
  def drawPart(penThickness,color,label,x,y,length, height)
  end
  def drawBoard(penThickness,color,label,x,y,length,height)
  end
end


#-----------------------------------------------------------------------------
# html renderer to write a java script to draw
# using wz_jsgraphics.js
#-----------------------------------------------------------------------------
class HtmlLayoutRenderer  < LayoutRenderer
  def header(x,y)
    # we'll need to add some things to the heading later so remember where it was
    html = ""
    #html = html + htmlHeader
    #html = html + scriptHeader
    html = html + functionHeader
    return html
  end

  def displayText(x,y,string)
    html = ""
    html += setColor("#7d7d4f")
    html += " cutlistLayout.setFont(\"verdana\",\"12px\",Font.BOLD);"  
    html += " cutlistLayout.drawString(\"#{string}\",#{x},#{y});"
    return html
  end

  def displayBoardFeet(string)
    html = ""
    x = @headingX + 500
    y = @headingY + 10
    html += displayText(x,y,string)
    return html
  end
  
  def displayEfficiency(string)
    html = ""
    x = @headingX + 330
    y = @headingY + 10
    html += displayText(x,y,string)
    return html
  end
  
  # input numbers startx,starty,length,height must be integer not float
  # color must be a color name string( eg 'blue' ) or hex string( eg #0000ff)
  # x,y is top left corner
  # This draws a filled rectangle using the color selected, with a black border and the text in the center
  def drawPart(penThickness,color,label,x,y,length, height)
    #label is placed in the approx center of the box
    labely = y + (height/2) -5
    html = ""
    html = html + setColor(color)
    html = html + " cutlistLayout.fillRect(#{x},#{y},#{length},#{height});"
    html = html + " cutlistLayout.setStroke(#{penThickness});"
    html = html + setColor("black")
    html = html + " cutlistLayout.setFont(\"arial\",\"10px\",Font.BOLD);"  
    # outline the rectangle using black
    html = html + " cutlistLayout.drawRect(#{x},#{y},#{length},#{height});"
    html = html + " cutlistLayout.drawStringRect(\"#{label}\",#{x},#{labely},#{length},\"center\");"
    return html
  end
  
  # a board starts off as a box, the background shade or pattern indicating it is unused
  # it has a label just above the board which is freeform text and can be used to indicate size or any notes
  # specific to the board
  # The color is the color of the background. All text is in black
  def drawBoard(penThickness,color,label,x,y,length,height)
    labely = y-20
    html = ""
    # put each board in its own cell so it becomes scrollable
    #html += "cutlistLayout.htm += \'<div style=\"position:relative;width:#{length+50}px;height:#{height+50}px;\"><table cellpadding=\"0\" cellspacing=\"0\"><tr><td>\';"
    html += setColor(color)
    html += " cutlistLayout.fillRect(#{x},#{y},#{length},#{height});"
    html += setColor("black")
    html += " cutlistLayout.setFont(\"arial\",\"10px\",Font.BOLD_ITALIC);" 
    html += hatch(x,y,length,height)
    html += " cutlistLayout.setStroke(#{penThickness});"
    html += " cutlistLayout.drawRect(#{x},#{y},#{length},#{height});"
    html += " cutlistLayout.drawStringRect(\"#{label}\",#{x},#{labely},#{length},\"left\");"
    #html += "cutlistLayout.htm += \'<\/td><\/tr><\/table><\/div>\';"
    #html += "cutlistLayout.paint();"
    return html
  end
  
  def hatch(x,y,length,height)
    html = ""
    html = html + " cutlistLayout.setStroke(Stroke.DOTTED);"
    lineSpacing = 20
    totalLines = ((length+height)/lineSpacing).floor
    # dim1 should always be the shorter of the two sides
    if height <= length
      dim1 = height
      dim2 = length
      horizontal=true
    else
      dim1 = length
      dim2 = height
      horizontal=false
    end
    #puts totalLines.to_s
    # y= -x+i*lineSpacing is the equation of the diagonals since positive y axis is towards the bottom of the screen
    1.upto(totalLines){ |i|
      if (i*lineSpacing <= dim1)
        startx=x
        endy=y
        starty=y+(x+i*lineSpacing) - startx
        endx=x+(y+i*lineSpacing) - endy
      elsif ( (i*lineSpacing > dim1) && (i*lineSpacing <= dim2) )
        if horizontal
          starty=y+height
          endy=y
          startx= x+ (y+i*lineSpacing) - starty
          endx=x+ (y+i*lineSpacing) - endy
        elsif
          startx=x
          endx=x+length
          endy=y+(x+i*lineSpacing)-endx
          starty=y+(x+i*lineSpacing)-startx
        end
      else
        starty=y+height
        endx=x+length
        startx=x+ (y+i*lineSpacing) - starty
        endy=y+ (x+i*lineSpacing) - endx
      end
      
      #draw the line from (startx,starty) to (endx,endy)
      html = html + "cutlistLayout.drawLine(#{startx},#{starty},#{endx},#{endy});"
    }
    return html
  end
  
  def pageHeading(x,y,string)
    html = ""
    # display the model name at the top of the page
    html = html + setColor("#7d7d4f")
    html = html + " cutlistLayout.setFont(\"verdana\",\"14px\",Font.BOLD);"  
    html = html + " cutlistLayout.drawString(\"#{string}\",#{x},#{y});"
    return html
  end
  
  def sectionHeading(x,y,type)
    # record where the heading ends up so we can add summaries to it later
    html = ""
    @headingX = x
    @headingY = y
    html = html + setColor("#7d7d4f")
    html = html + " cutlistLayout.setFont(\"verdana\",\"20px\",Font.BOLD);"  
    html = html + " cutlistLayout.drawString(\"Cutting Diagram  - #{type}\",#{x},#{y});"
    return html
  end
  
  def setColor(color)
    return " cutlistLayout.setColor(\"#{color}\");"
  end
  
  def footer
    html = ""
    html = html + functionFooter
    #html = html + scriptFooter
    return html
  end
    
  def functionHeader
    html = ""
    html = html +"function drawLayout()"
    html = html + "{"
    #html = html + "  cutlistLayout.setPrintable(true);"
    return html
  end
  
  def draw
    html = ""
    html = html + "  cutlistLayout.paint();"
    return html
  end
  
  def functionFooter
    html = ""
    html += draw
    html = html + "}"
    html = html +"var cutlistLayout = new jsGraphics(\"layoutDiv\");"
    #html = html +"drawLayout();"
    return html
  end
  
end

#-----------------------------------------------------------------------------
# svg renderer to generate svg format text output description of the graphics
# which can then be displayed using an scvg browser
#-----------------------------------------------------------------------------
class SvgLayoutRenderer  < HtmlLayoutRenderer
  def functionHeader
    html = ""
    html += "<?xml version=\"1.0\" standalone=\"no\"?>\n"
    html += "<!DOCTYPE svg PUBLIC \"-//W3C//DTD SVG 1.1//EN\" \n"
    html += "  \"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd\">\n"
    html += " \n"
    html += "<svg width=\"100%\" height=\"100%\" version=\"1.1\" \n"
    html += "  xmlns=\"http://www.w3.org/2000/svg\">\n"
    return html
  end
  
  def text(x,y,string,color,font,fontSize,fontWeight,fontStyle,textAnchor)
    html = ""
    html += "<text"
    html += " x=\"#{x}\" y=\"#{y}\" fill=\"#{color}\""
    html += " font-family=\"#{font}\" font-size=\"#{fontSize}\" font-weight=\"#{fontWeight}\" font-style=\"#{fontStyle}\""
    html += " text-anchor=\"#{textAnchor}\">"
    html += "#{string}</text>\n"
    return html
  end
  
  def displayText(x,y,string)
    html = ""
    html += text(x,y,string,"#7d7d4f","Verdana",12,"bold","normal","start")
    return html
  end
  
  def pageHeading(x,y,string)
    html = ""
    # Always display the model name at the top
    html += text(x,y,string,"#7d7d4f","Verdana",14,"bold","normal","start")
    return html
  end
  
  def sectionHeading(x,y,type)
    # record where our cutting diagram heading ends up so we can come back and add efficiency summaries to it later
    html = ""
    @headingX = x
    @headingY = y
    y += 10
    # Display the cutting diagram header and board type
    string = "Cutting Diagram  - #{type}"
    html += text(x,y,string,"#7d7d4f","Verdana",20,"bold","normal","start")
    return html
  end
  
  def drawPart(penThickness,color,label,x,y,length, height)
    #label is placed in the approx center of the box
    labely = y + (height/2) +2
    labelx = x + (length/2) + 2
    html = ""
    html = ""
    html += "<rect"
    html += " x=\"#{x}\" y=\"#{y}\" fill=\"#{color}\" width=\"#{length}\" height=\"#{height}\" stroke=\"black\" stroke-width=\"#{penThickness}\">"
    html += "</rect>\n"
    html += text(labelx,labely,label,"black","arial",8,"bold","normal","middle")
    return html
  end
    
  def drawBoard(penThickness,color,label,x,y,length,height)
    labely = y-10
    html = ""
    html += "<rect"
    html += " x=\"#{x}\" y=\"#{y}\" fill=\"#{color}\" width=\"#{length}\" height=\"#{height}\" stroke=\"black\" stroke-width=\"#{penThickness}\">"
    html += "</rect>\n"
    html += text(x,labely,label,"black","arial",10,"bold","italic","start")
    html += hatch(x,y,length,height)
    return html
  end
  
  def hatch(x,y,length,height)
    html = ""
    lineSpacing = 20
    totalLines = ((length+height)/lineSpacing).floor
    # dim1 should always be the shorter of the two sides
    if height <= length
      dim1 = height
      dim2 = length
      horizontal=true
    else
      dim1 = length
      dim2 = height
      horizontal=false
    end
    #puts totalLines.to_s
    # y= -x+i*lineSpacing is the equation of the diagonals since positive y axis is towards the bottom of the screen
    1.upto(totalLines){ |i|
      if (i*lineSpacing <= dim1)
        startx=x
        endy=y
        starty=y+(x+i*lineSpacing) - startx
        endx=x+(y+i*lineSpacing) - endy
      elsif ( (i*lineSpacing > dim1) && (i*lineSpacing <= dim2) )
        if horizontal
          starty=y+height
          endy=y
          startx= x+ (y+i*lineSpacing) - starty
          endx=x+ (y+i*lineSpacing) - endy
        elsif
          startx=x
          endx=x+length
          endy=y+(x+i*lineSpacing)-endx
          starty=y+(x+i*lineSpacing)-startx
        end
      else
        starty=y+height
        endx=x+length
        startx=x+ (y+i*lineSpacing) - starty
        endy=y+ (x+i*lineSpacing) - endx
      end
      
      #draw the line from (startx,starty) to (endx,endy)
    html += "<line"
    html += " x1=\"#{startx}\" y1=\"#{starty}\" x2=\"#{endx}\" y2=\"#{endy}\" "
    html += "stroke=\"grey\" stroke-width=\"1\" />\n"
    }
    return html
  end
      
  def functionFooter
    html = ""
    html = html +"</svg>\n"
    return html
  end
end


#########################
# FileRenderer          #
#########################
class FileRenderer < Renderer
  
  def initialize
    # call any intialisation required for the bsae class
    super
    locale = Sketchup.get_locale
    puts "locale=" + locale
    if ( locale == 'en-US' ||
         locale == 'en-GB' )
      # The following line, if uncommented, uses tab characters instead of ',' as the delimiter
      #@delimiter = "\ci"
      @delimiter = ","
    else
      @delimiter = ";"
    end
  end

  def getTitle(title)
    # don't print a blank line if there is no title - this is important for CutList Plus which looks for headings
    # in the first line
    txt = ""
    txt = title+"\n" if title != ""
    return txt
  end ## end getTitle

  def getHeaderRow(headers)
    txt = ""
    for h in headers
      txt = txt+h
      txt = txt+@delimiter 
    end ## end for
    txt = txt+"\n"
    return txt
  end ## end getHeaderRow

  def getFooterRow()
    return ""
  end ## end getFooterRow

  def getRow(columns)
    txt = ""
    for c in columns
      txt = txt+c
      txt = txt+@delimiter 
    end ## end for
    txt = txt+"\n"
    return txt
  end ## end getRow

  def getAmount(amount)
    return amount+"\n"
  end ## end getArea

  def getBlankLine()
   return "\n"
  end ## end getBlankLine

end ## end class FileRenderer
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# create a Reporter instance and run a fixed configuration, producing only the file outputs
#def fixed_configuration_generator
#  reporter = Reporter.new
#  reporter.sketchup5Init
#end

# create a Reporter instance that prompts for an interactive configuration, producing the requested output formats
def interactive_generator
  webGui = WebGui.new
  webGui.start
end

# Add things to the Plugins menu
# Add CutList main entry 
# "Cut List" offers an html gui to select options and produce html and/or file output 
if( not file_loaded?("CutListAndMaterials.rb") )
  add_separator_to_menu("Plugins")
  
  #plugins_menu = UI.menu("Plugins").add_submenu("Cut List")
  plugins_menu = UI.menu("Plugins")
  
  plugins_menu.add_item("Cut List") { interactive_generator }
  # no longer supported
  #plugins_menu.add_item("SU5") { fixed_configuration_generator }
end 

file_loaded("CutListAndMaterials.rb")
#-----------------------------------------------------------------------------
